; Canned comparative reports       -*- mode: scheme; mode: font-lock -*-
; 2000-09-01 / lth

(require 'control)
(require 'format)

; Common code

(define (all-benchmarks-gnuplot-report file-or-port results)
  (call-with-output-medium file-or-port
    (lambda (out)
      (for-each (lambda (x)
                  (format out "~a ~a~%" (car x) (cdr x)))
                results)
      (format #t "~a: ~a data points total.~%"
              file-or-port
              (sum cdr results)))))

(define (all-sc-canonical-data)
  (list (sc-canonical-data)))

(define (all-2gen-canonical-data)
  (list (2gen-canonical-data)))

(define (all-3gen-canonical-data)
  (list (3gen-canonical-data)))

(define (all-gen-canonical-data)  
  (list (2gen-canonical-data)
        (3gen-canonical-data)))

(define (all-dof-canonical-data)
  (list (1dof-canonical-data)
        (2dof-canonical-data)
        (3dof-canonical-data)
        (4dof-canonical-data)))

(define (all-1dof-canonical-data . rest) ; optional arg: list of benchmarks
  (list (apply 1dof-canonical-data rest)))

(define (all-2dof-canonical-data . rest)
  (list (apply 2dof-canonical-data rest)))

(define (all-3dof-canonical-data . rest)
  (list (apply 3dof-canonical-data rest)))

(define (all-4dof-canonical-data . rest)
  (list (apply 4dof-canonical-data rest)))

(define (all-rof-canonical-data)
  (list (rof-canonical-data)))

(define (report-box-and-line-by-benchmark comparer file d1 d2 . rest)
  (call-with-output-file file
    (lambda (out)
      (apply 
       box-and-line-diagrams 
       `(,out
         ,(extract-benchmark-trend-data (comparer d1 d2))
         ,@(if (not (null? rest))
               rest
               '(#f #f #f #f))
         down 630)))))

; SC vs GEN

(define (report-sc-and-gen-by-cpu-time file)
  (all-benchmarks-gnuplot-report 
   file 
   (compare-all-by-cpu-time (all-sc-canonical-data) (all-gen-canonical-data))))

(define (report-sc-and-gen-by-cpu-gc-time file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-gc-time (all-sc-canonical-data) (all-gen-canonical-data))))

(define (report-sc-and-gen-by-cpu-mutator-time file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-mutator-time (all-sc-canonical-data) (all-gen-canonical-data))))

(define (report-sc-and-gen-by-space file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-space (all-sc-canonical-data) (all-gen-canonical-data))))

(define (report-sc-and-gen-by-benchmark comparer file . rest)
  (apply report-box-and-line-by-benchmark 
         comparer
         file
         (all-sc-canonical-data)
         (all-2gen-canonical-data)    ; was: (all-gen-canonical-data)
         rest))

(define (report-sc-and-gen-by-benchmark-and-cpu-time file)
  (report-sc-and-gen-by-benchmark compare-benchmarks-by-cpu-time file
                                  -100 150 25 -100))

(define (report-sc-and-gen-by-benchmark-and-cpu-gc-time file)
  (report-sc-and-gen-by-benchmark compare-benchmarks-by-cpu-gc-time file
                                  -350 950 100 -300))

(define (report-sc-and-gen-by-benchmark-and-cpu-mutator-time file)
  (report-sc-and-gen-by-benchmark compare-benchmarks-by-cpu-mutator-time file))

(define (report-sc-and-gen-by-benchmark-and-space file)
  (report-sc-and-gen-by-benchmark compare-benchmarks-by-space file))

(define (report-sc-and-2gen-by-benchmark-space-time prefix)
  (report-by-benchmark-space-time prefix
                                  (sc-canonical-data) 
                                  (2gen-canonical-data)))

(define (report-sc-and-3gen-by-benchmark-space-time prefix)
  (report-by-benchmark-space-time prefix
                                  (sc-canonical-data) 
                                  (3gen-canonical-data)))

; NOTE! that report-sc-and-gen-by-benchmark uses 2GEN canonical data only.

(define (canned-sc-vs-2gen-plots)
  (display "SC vs 2GEN plots")
  (newline)
  (let ((dir (lambda (x) (string-append "plots/sc-vs-2gen/" x))))
    (report-sc-and-gen-by-cpu-time (dir "cpu-time.data"))
    (report-sc-and-gen-by-cpu-gc-time (dir "cpu-gc-time.data"))
    (report-sc-and-gen-by-cpu-mutator-time (dir "cpu-mutator-time.data"))
    (report-sc-and-gen-by-space (dir "space.data"))
    (report-sc-and-gen-by-benchmark-and-cpu-time (dir "bm-cpu-time.tex"))
    (report-sc-and-gen-by-benchmark-and-cpu-gc-time (dir "bm-cpu-gc-time.tex"))
    (report-sc-and-gen-by-benchmark-and-cpu-mutator-time 
     (dir "bm-cpu-mutator-time.tex"))
    (report-sc-and-gen-by-benchmark-and-space (dir "bm-space.tex"))
    #t))

(define (canned-sc-vs-gen-plots)
  (sc-vs-gen-consistency)
  (report-sc-and-2gen-by-benchmark-space-time "plots/sc-vs-gen/2st-")
  (report-sc-and-3gen-by-benchmark-space-time "plots/sc-vs-gen/3st-"))

; A weak notion of consistency: data sets are of the same size.
; A better notion of consistency looks at comparable runs.

(define (sc-vs-gen-consistency)
  (test-consistency "sc-vs-gen-consistency"
                    (sc-canonical-data)
                    (2gen-canonical-data)
                    (3gen-canonical-data)))

(define (gen-vs-dof-consistency)
  (test-consistency "gen-vs-dof-consistency"
                    (1dof-canonical-data) 
                    (2dof-canonical-data) 
                    (3dof-canonical-data) 
                    (4dof-canonical-data) 
                    (2gen-canonical-data)
                    (3gen-canonical-data)))

(define (gen-vs-rof-consistency)
  (test-consistency "gen-vs-rof-consistency"
                    (rof-canonical-data) 
                    (2gen-canonical-data)
                    (3gen-canonical-data)))

(define (test-consistency name . data)

  (define (pair-up l)
    (cond ((null? l) '())
          ((null? (cdr l)) 
           (error "pair-up: only one element: " l))
          (else
           (let loop ((x (car l)) (l (cdr l)) (r '()))
             (if (null? l)
                 (reverse r)
                 (loop (car l) (cdr l) (cons (cons x (car l)) r)))))))
  (apply
   for-each-benchmark-result* 
   (lambda (n . data)
     (if (not (every? (lambda (x)
                        (comparable-runs? (car x) (cdr x)))
                      (pair-up data)))
         (error name ": Bogus: "
                (map (lambda (x) (assq 'system-info x)) data))))
   data))

; 2GEN vs 3GEN

(define (report-2gen-and-3gen-by-cpu-time file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-time (all-2gen-canonical-data) (all-3gen-canonical-data))))

(define (report-2gen-and-3gen-by-cpu-gc-time file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-gc-time (all-2gen-canonical-data) (all-3gen-canonical-data))))

(define (report-2gen-and-3gen-by-cpu-mutator-time file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-mutator-time (all-2gen-canonical-data) (all-3gen-canonical-data))))

(define (report-2gen-and-3gen-by-space file)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-space (all-2gen-canonical-data) (all-3gen-canonical-data))))

(define (report-2gen-and-3gen-by-benchmark comparer file . rest)
  (apply report-box-and-line-by-benchmark
         comparer
         file
         (all-2gen-canonical-data)
         (all-3gen-canonical-data)
         rest))

(define (report-2gen-and-3gen-by-benchmark-and-cpu-time file)
  (report-2gen-and-3gen-by-benchmark compare-benchmarks-by-cpu-time
                                     file))

(define (report-2gen-and-3gen-by-benchmark-and-cpu-gc-time file)
  (report-2gen-and-3gen-by-benchmark compare-benchmarks-by-cpu-gc-time file))

(define (report-2gen-and-3gen-by-benchmark-and-cpu-mutator-time file)
  (report-2gen-and-3gen-by-benchmark compare-benchmarks-by-cpu-mutator-time 
                                     file))

(define (report-2gen-and-3gen-by-benchmark-and-space file)
  (report-2gen-and-3gen-by-benchmark compare-benchmarks-by-space file))

(define (report-2gen-and-3gen-by-benchmark-space-time prefix)
  (report-by-benchmark-space-time prefix
                                  (2gen-canonical-data)
                                  (3gen-canonical-data)))

(define (canned-2gen-vs-3gen-plots)
  (display "2GEN vs 3GEN plots")
  (newline)

  (let ((dir (lambda (x) (string-append "plots/2gen-vs-3gen/" x))))
    (report-2gen-and-3gen-by-cpu-time (dir "cpu-time.data"))
    (report-2gen-and-3gen-by-cpu-gc-time (dir "cpu-gc-time.data"))
    (report-2gen-and-3gen-by-cpu-mutator-time (dir "cpu-mutator-time.data"))
    (report-2gen-and-3gen-by-space (dir "space.data"))
    (report-2gen-and-3gen-by-benchmark-and-cpu-time (dir "bm-cpu-time.tex"))
    (report-2gen-and-3gen-by-benchmark-and-cpu-gc-time 
     (dir "bm-cpu-gc-time.tex"))
    (report-2gen-and-3gen-by-benchmark-and-cpu-mutator-time
     (dir "bm-cpu-mutator-time.tex"))
    (report-2gen-and-3gen-by-benchmark-and-space (dir "bm-space.tex"))
    (report-2gen-and-3gen-by-benchmark-space-time (dir "st-"))
    #t))


;----------------------------------------------------------------------
;
; GEN vs DOF

(define (report-gen-and-dof-by-cpu-time file gen-data dof-data)
  (all-benchmarks-gnuplot-report
   file
   (compare-all-by-cpu-time gen-data dof-data)))

(define (report-gen-and-dof-by-cpu-gc-time file gen-data dof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-gc-time gen-data dof-data)))

(define (report-gen-and-dof-by-cpu-mutator-time file gen-data dof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-mutator-time gen-data dof-data)))

(define (report-gen-and-dof-by-space file gen-data dof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-space gen-data dof-data)))

(define (canned-gen-vs-dof-plots:overall)
  (let ((gen-data  (all-gen-canonical-data))
        (2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (dof-data  (all-dof-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data))
        (dir (lambda xs (apply string-append "plots/gen-vs-dof/" xs))))

    ; Best of GEN vs best of best of DOF 
    (report-gen-and-dof-by-cpu-time (dir "cpu-time.data") gen-data dof-data)
    (report-gen-and-dof-by-cpu-gc-time (dir "cpu-gc-time.data")
                                       gen-data dof-data)
    (report-gen-and-dof-by-cpu-mutator-time (dir "cpu-mutator-time.data")
                                            gen-data dof-data)
    (report-gen-and-dof-by-space (dir "space.data") gen-data dof-data)
  
    ; Best of mGEN vs best of nDOF across all benchmarks

    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gen))
      (do ((dofcnt '(1 2 3 4) (cdr dofcnt))
           (dof    (list 1dof-data 2dof-data 3dof-data 4dof-data) (cdr dof)))
          ((null? dof))
        (let ((name (string-append (number->string (car gencnt))
                                   "gen-"
                                   (number->string (car dofcnt))
                                   "dof.data")))
          (report-gen-and-dof-by-cpu-time 
           (dir "cpu-time-" name) (car gen) (car dof))
          (report-gen-and-dof-by-cpu-gc-time
           (dir "cpu-gc-time-" name) (car gen) (car dof))
          (report-gen-and-dof-by-cpu-mutator-time 
           (dir "cpu-mutator-time-" name) (car gen) (car dof))
          (report-gen-and-dof-by-space 
           (dir "space-" name) (car gen) (car dof)))))

    #t))

(define (gen-vs-dof-space-time-plots prefix)
  (do ((gencnt '(2 3) (cdr gencnt))
       (gen    (list (2gen-canonical-data) (3gen-canonical-data))
               (cdr gen)))
      ((null? gen))
    (do ((dofcnt '(1 2 3 4) (cdr dofcnt))
         (dof    (list (1dof-canonical-data) (2dof-canonical-data)
                       (3dof-canonical-data) (4dof-canonical-data))
                 (cdr dof)))
        ((null? dof))
      (let ((name (string-append prefix
                                 (number->string (car gencnt))
                                 "gen-"
                                 (number->string (car dofcnt))
                                 "dof-")))
        (display (string-append name "..."))
        (newline)
        (report-by-benchmark-space-time name (car gen) (car dof)))))
  #t)

(define (gen-vs-dof-avg-advantage-tables)
  (load-gen-results)
  (load-dof-results)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data))
        (dir (lambda xs (apply string-append "tables/" xs))))

     
    (define (prelude out)
      (format out "\\begin{tabular}{|l|rrrrr|rrrrr|}~%\\hline~%")
      (format out " & \\multicolumn{5}{c|}{Total time}")
      (format out " & \\multicolumn{5}{c|}{GC time} \\\\~%")
      (format out " & Avg & \\multicolumn{4}{c|}{Range}")
      (format out " & Avg & \\multicolumn{4}{c|}{Range} \\\\~%\\hline~%"))

    (define (postlude out)
      (format out "\\hline~%\\end{tabular}~%"))

    (define d (decimals 1))

    ; negative advantage is advantage to GEN.

    (define (latex-table start? i out results)
      (let* ((results (sort results (lambda (a b) (< (car a) (car b)))))
             (min (caar results))
             (max (caar (last-pair results)))
             (avg (weighted-mean results car cdr))
             (iq  (interquartile 
                   (lambda (x) 
                     (* (+ (abs min) (car x)) (cdr x))) 
                   results))
             (iq-min (caar iq))
             (iq-max (caar (last-pair iq))))
        (format out " ~a & ~a & ~a & ~a & ~a & ~a ~a" 
                (if start? (format #f "~aDOF" i) "")
                (d avg) min iq-min iq-max max
                (if (not start?) (format #f "\\\\~%") ""))))

    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gen))
      (let* ((name (string-append "avg-adv-"
                                  (number->string (car gencnt))
                                  "gen.tex"))
             (out (open-output-file (dir name))))
        (display (dir name)) 
        (newline)
        (prelude out)
        (do ((dofcnt '(1 2 3 4) (cdr dofcnt))
             (dof    (list 1dof-data 2dof-data 3dof-data 4dof-data) (cdr dof)))
            ((null? dof))
          (latex-table #t
           (car dofcnt) out (compare-all-by-cpu-time (car gen) (car dof)))
          (latex-table #f
           (car dofcnt) out (compare-all-by-cpu-gc-time (car gen) (car dof))))
        (postlude out)
        (close-output-port out)))
    #t))

(define (gen-vs-dof-count-wins-table)

  (define colsums '())

  (define (prelude out)
    (format out "\\begin{tabular}{|l|rrrr|rrrr|}~%\\hline~%")
    (format out " & \\multicolumn{4}{c|}{Total time}")
    (format out " & \\multicolumn{4}{c|}{GC time} \\\\~%")
    (format out " & 1DOF & 2DOF & 3DOF & 4DOF")
    (format out " & 1DOF & 2DOF & 3DOF & 4DOF \\\\~%\\hline~%")
    (set! colsums (make-list 8 0)))

  (define (postlude out)
    (format out "\\hline~%\\end{tabular}~%"))

  (define (body out bm total-results gc-results)
    (format out "~20a " (format-benchmark-name bm))
    (apply format out " & ~3@a & ~3@a & ~3@a & ~3@a" total-results)
    (apply format out " & ~3@a & ~3@a & ~3@a & ~3@a \\\\~%" gc-results)
    (set! colsums (map + colsums (append total-results gc-results))))

  (define (tally-victories res bm)
    (let ((bm-results (cdr (assoc bm res))))
      (inexact->exact
       (round 
        (* 100.0 (/ (sum (lambda (x) (if (>= (car x) 0) (cdr x) 0)) 
                         bm-results)
                    (sum (lambda (x) (cdr x)) bm-results)))))))

  (load-dof-results)
  (load-gen-results)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data)))
    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gencnt))
      (let ((fn (string-append "tables/gen-vs-dof-count"
                               (number->string (car gencnt))
                               ".tex")))
        (display fn)
        (newline)
        (call-with-output-file fn
          (lambda (out)
            (prelude out)
            (let* ((gen-data (car gen))
                   (res-t 
                    (list
                     (compare-benchmarks-by-cpu-time gen-data 1dof-data)
                     (compare-benchmarks-by-cpu-time gen-data 2dof-data)
                     (compare-benchmarks-by-cpu-time gen-data 3dof-data)
                     (compare-benchmarks-by-cpu-time gen-data 4dof-data)))
                   (res-gc 
                    (list
                     (compare-benchmarks-by-cpu-gc-time gen-data 1dof-data)
                     (compare-benchmarks-by-cpu-gc-time gen-data 2dof-data)
                     (compare-benchmarks-by-cpu-gc-time gen-data 3dof-data)
                     (compare-benchmarks-by-cpu-gc-time gen-data 4dof-data))))
              (do ((bms benchmark-names (cdr bms)))
                  ((null? bms))
                (let ((bm (car bms)))
                  (body out bm 
                        (map (lambda (r) (tally-victories r bm)) res-t)
                        (map (lambda (r) (tally-victories r bm)) res-gc))))
              (format out "\\hline~%~20a" "Sum")
              (for-each (lambda (x) (format out " & ~3@a" x))
                        colsums)
              (format out "\\\\~%"))
            (postlude out)))))))

; Average advantage and number of wins

(define (gen-vs-dof-space-table)

  (define colsums '())

  (define (prelude out)
    (format out "\\begin{tabular}{|l|rrrr|}~%\\hline~%")
    (format out " & \\multicolumn{4}{c|}{Space} \\\\~%")
    (format out " & 1DOF & 2DOF & 3DOF & 4DOF \\\\~%\\hline~%")
    (set! colsums (make-list 4 0)))

  (define (postlude out)
    (format out "\\hline~%\\end{tabular}~%"))

  (define (body out bm results)
    (format out "~20a " (format-benchmark-name bm))
    (apply format out " & ~3@a & ~3@a & ~3@a & ~3@a \\\\~%" results)
    (set! colsums (map + colsums results)))

  (define (tally-victories res bm)
    (let ((bm-results (cdr (assoc bm res))))
      (inexact->exact
       (round 
        (* 100.0 (/ (sum (lambda (x) (if (>= (car x) 0) (cdr x) 0)) 
                         bm-results)
                    (sum (lambda (x) (cdr x)) bm-results)))))))

  (load-dof-results)
  (load-gen-results)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data)))
    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gencnt))
      (let ((fn (string-append "tables/gen-vs-dof-count-space"
                               (number->string (car gencnt))
                               ".tex")))
        (display fn)
        (newline)
        (call-with-output-file fn
          (lambda (out)
            (prelude out)
            (let* ((gen-data (car gen))
                   (res
                    (list (compare-benchmarks-by-space gen-data 1dof-data)
                          (compare-benchmarks-by-space gen-data 2dof-data)
                          (compare-benchmarks-by-space gen-data 3dof-data)
                          (compare-benchmarks-by-space gen-data 4dof-data))))
              (do ((bms benchmark-names (cdr bms)))
                  ((null? bms))
                (let ((bm (car bms)))
                  (body out bm (map (lambda (r) (tally-victories r bm)) res))))
              (format out "\\hline~%~20a" "Sum")
              (for-each (lambda (x) (format out " & ~3@a" x))
                        colsums)
              (format out "\\\\~%"))
            (postlude out)))))))

; More hackery!
(define (gen-vs-dof-total-space-table)

  (define colsums '())

  (define (prelude out)
    (format out "\\begin{tabular}{|l|rrrr|}~%\\hline~%")
    (format out " & \\multicolumn{4}{c|}{Space} \\\\~%")
    (format out " & 1DOF & 2DOF & 3DOF & 4DOF \\\\~%\\hline~%")
    (set! colsums (make-list 4 0)))

  (define (postlude out)
    (format out "\\hline~%\\end{tabular}~%"))

  (define (body out bm results)
    (format out "~20a " (format-benchmark-name bm))
    (apply format out " & ~3@a & ~3@a & ~3@a & ~3@a \\\\~%" results)
    (set! colsums (map + colsums results)))

  (define (tally-victories res bm)
    (let ((bm-results (cdr (assoc bm res))))
      (inexact->exact
       (round 
        (* 100.0 (/ (sum (lambda (x) (if (>= (car x) 0) (cdr x) 0)) 
                         bm-results)
                    (sum (lambda (x) (cdr x)) bm-results)))))))

  (load-dof-results)
  (load-gen-results)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data)))
    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gencnt))
      (let ((fn (string-append "tables/gen-vs-dof-count-totalspace"
                               (number->string (car gencnt))
                               ".tex")))
        (display fn)
        (newline)
        (call-with-output-file fn
          (lambda (out)
            (prelude out)
            (let* ((gen-data (car gen))
                   (res
                    (list (compare-benchmarks-by-total-space
                           gen-data 1dof-data)
                          (compare-benchmarks-by-total-space 
                           gen-data 2dof-data)
                          (compare-benchmarks-by-total-space
                           gen-data 3dof-data)
                          (compare-benchmarks-by-total-space
                           gen-data 4dof-data))))
              (do ((bms benchmark-names (cdr bms)))
                  ((null? bms))
                (let ((bm (car bms)))
                  (body out bm (map (lambda (r) (tally-victories r bm)) res))))
              (format out "\\hline~%~20a" "Sum")
              (for-each (lambda (x) (format out " & ~3@a" x))
                        colsums)
              (format out "\\\\~%"))
            (postlude out)))))))

(define (report-gen-vs-dof-qualitative)

  (define (process out)
    (let ((res (gen-vs-dof-qualitative qualitative-test)))
      (format out "\\begin{tabular}{|l|rr|rr|rr|rr|}~%\\hline~%")
      (do ((i 1 (+ i 1)))
          ((> i 4))
        (format out " & \\multicolumn{2}{c|}{~aDOF}" i))
      (format out "\\\\~%")
      (format out " & GEN & DOF & GEN & DOF & GEN & DOF & GEN & DOF\\\\~%")
      (format out "\\hline~%")
      (do ((i 2 (+ i 1))
           (res res (cdr res)))
          ((> i 3))
        (format out "~aGEN " i)
        (do ((i 0 (+ i 1))
             (r (list-ref (car res) i) (list-ref (car res) i)))
            ((= i 4))
          (let ((n (vector-ref r 5)))
            (format out " & ~4,1f & ~4,1f" 
                    (* 100 (/ (+ (vector-ref r 0) (vector-ref r 1)) n))
                    (* 100 (/ (+ (vector-ref r 2) (vector-ref r 3)) n)))))
        (format out "\\\\~%"))
      (format out "\\hline~%")
      (format out "\\end{tabular}~%")))
      
  (process (current-output-port)))

(define (report-gen-vs-dof-qualitative-by-benchmark)

  (define (process out)
    (let ((res (gen-vs-dof-qualitative 
                qualitative-test-by-benchmark)))
      (format out "\\begin{tabular}{|ll|rr|rr|rr|rr|}~%\\hline~%")
      (do ((i 1 (+ i 1)))
          ((> i 4))
        (format out " && \\multicolumn{2}{c|}{~aDOF}" i))
      (format out "\\\\~%")
      (format out " && GEN & DOF & GEN & DOF & GEN & DOF & GEN & DOF\\\\~%")
      (do ((i 2 (+ i 1))
           (res res (cdr res))
           (name #f))
          ((> i 3))
        (format out "\\hline~%")
        (set! name (format #f "~aGEN" i))
        (for-each 
         (lambda (bm)
           (format out " ~a & ~20a" name (format-benchmark-name bm))
           (set! name "    ")
           (do ((j 0 (+ j 1)))
               ((= j 4))
             (let* ((r (list-ref (car res) j))
                    (v (cadr (assoc bm r)))
                    (n (vector-ref v 5)))
               (format out " & ~5,1f & ~5,1f" 
                       (* 100 (/ (+ (vector-ref v 0) (vector-ref v 1)) n))
                       (* 100 (/ (+ (vector-ref v 2) (vector-ref v 3)) n)))))
           (format out "\\\\~%"))
         benchmark-names))
      (format out "\\hline~%")
      (format out "\\end{tabular}~%")))
      
  (process (current-output-port)))

; results are 
;  (2gen-results 3gen-results)
; ngen-results are
;  (#(a b c d e f) ...)
; where the vectors are the results of comparing the gen
; collector against 1dof .. 4dof (in order) and
;  a+b = number of wins for the generational system
;  c+d = number of wins for the dof system
;  e   = number of unclassified compared runs
;  f   = total number of compared runs

(define (gen-vs-dof-qualitative subprogram)

  (load-gen-results)
  (load-dof-results)
  (let ((2gen-data (2gen-canonical-data))
        (3gen-data (3gen-canonical-data))
        (1dof-data (1dof-canonical-data))
        (2dof-data (2dof-canonical-data))
        (3dof-data (3dof-canonical-data))
        (4dof-data (4dof-canonical-data)))
    
    (map
     (lambda (gen)
       (map
        (lambda (dof)
          (subprogram gen dof))
        (list 1dof-data 2dof-data 3dof-data 4dof-data)))
     (list 2gen-data 3gen-data))))

; Compare costs of copying both for collection and promotion.
; These use a single set of 2GEN and 4DOF runs, though the data
; could be used to construct separate tables for all nDOF.
; There is little point in doing so -- the cost should not vary
; greatly across the collectors, _except as influenced by 
; remembered set scanning_.  So be careful about the 
; interpretation.

(define (gen-vs-dof-copy-cost-report . rest)

  (define (bm<? a b)
    (benchmark<? (car a) (car b)))

  (define d (decimals 1))
  (define d2 (decimals 2))

  (define (doit out)
    (let ((gen (sort (gen-promotion/collection-cost) bm<?))
          (dof (sort (dof-promotion/collection-cost) bm<?)))
      (format out "\\begin{tabular}{|l|rrr|rrr|}~%")
      (format out "\\hline~%")
      (format out " & \\multicolumn{3}{c|}{Promote}")
      (format out " & \\multicolumn{3}{c|}{Collect} \\\\~%")
      (format out " & 2GEN & 4DOF & DOF/GEN")
      (format out " & 2GEN & 4DOF & DOF/GEN \\\\~%")
      (format out "\\hline~%")
      (for-each (lambda (gen dof)
                  (format out "~20a " (format-benchmark-name (car gen)))
                  (format out "& ~4@a & ~4@a & ~5@a "
                          (d (cadr gen)) (d (cadr dof)) 
                          (d2 (/ (cadr dof) (cadr gen))))
                  (format out "& ~4@a & ~4@a & ~5@a "
                          (d (caddr gen)) (d (caddr dof)) 
                          (d2 (/ (caddr dof) (caddr gen))))
                  (format out "\\\\~%"))
                gen dof)
      (format out "\\hline~%")
      (format out "\\end{tabular}~%")))

  (if (null? rest)
      (doit (current-output-port))
      (call-with-output-file (car rest) doit)))


(define (canned-gen-vs-dof-plots:by-benchmark)
  (let ((gen-data  (all-gen-canonical-data))
        (2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (dof-data  (all-dof-canonical-data))
        (1dof-data (all-1dof-canonical-data))
        (2dof-data (all-2dof-canonical-data))
        (3dof-data (all-3dof-canonical-data))
        (4dof-data (all-4dof-canonical-data))
        (dir (lambda xs (apply string-append "plots/gen-vs-dof/" xs))))

    ; Best of mGEN vs best of nDOF by benchmark

    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gen))
      (do ((dofcnt '(1 2 3 4) (cdr dofcnt))
           (dof    (list 1dof-data 2dof-data 3dof-data 4dof-data) (cdr dof)))
          ((null? dof))
        (do ((cmp (list compare-benchmarks-by-cpu-time
                        compare-benchmarks-by-cpu-gc-time
                        compare-benchmarks-by-cpu-mutator-time
                        compare-benchmarks-by-space)
                  (cdr cmp))
             (name '("cpu-time" "cpu-gc-time" "cpu-mutator-time" "space")
                   (cdr name)))
            ((null? cmp))
          (let ((name (string-append "bm-"
                                     (car name)
                                     "-"
                                     (number->string (car gencnt))
                                     "gen-"
                                     (number->string (car dofcnt))
                                     "dof.tex")))
            (display (dir name))
            (display "...")
            (newline)
            (report-box-and-line-by-benchmark
             (car cmp)
             (dir name)
             (car gen)
             (car dof))))))

    #t))

(define (canned-gen-vs-dof-plots)
  (display "GEN vs DOF plots")
  (newline)

  (canned-gen-vs-dof-plots:overall)
  (canned-gen-vs-dof-plots:by-benchmark)
  (gen-vs-dof-space-time-plots "plots/gen-vs-dof/st-")
  #t)


; Supporting data: heap space usage advantage, GEN vs DOF, simple
; frequency count.  Clever or stupid, this one just grubs through the
; data files produced by earlier comparative plots!

(define (gen-vs-dof-heap-space-counts)

  (define files
    '("plots/gen-vs-dof/space-2gen-1dof.data"
      "plots/gen-vs-dof/space-2gen-2dof.data"
      "plots/gen-vs-dof/space-2gen-3dof.data"
      "plots/gen-vs-dof/space-2gen-4dof.data"))

  (define (display-data r)
    (let ((out (current-output-port)))
      (format out "\\begin{tabular}{|l|rr|}~%\\hline~%")
      (format out "  & \\multicolumn{2}{c|}{Advantage (\\%)}\\\\~%")
      (format out "  &    GEN best & DOF best \\\\~%\\hline~%")
      (do ((r r (cdr r)))
          ((null? r))
        (apply format out "~aDOF & ~a & ~a \\\\~%" 
               (map (compose inexact->exact round) (car r))))
      (format out "\\hline~%")
      (format out "\\end{tabular}~%")))

  (gen-vs-*-heap-space-counts files display-data))

(define (gen-vs-rof-heap-space-counts . rest)

  (define files
    '("plots/gen-vs-rof/space-2gen-rof.data"))

  (define (doit out)
    (gen-vs-*-heap-space-counts files (lambda (r)
                                        (pretty-print r out))))

  (load-gen-results)
  (load-rof-results)
  (if (null? rest)
      (doit (current-output-port))
      (call-with-output-medium (car rest) doit)))

(define (gen-vs-*-heap-space-counts files display-data)

  (define (percent n)
    (round (* n 100.0)))

  (do ((n   1     (+ n 1))
       (f   files (cdr f))
       (res '()))
      ((null? f)
       (display-data (reverse res)))
    (call-with-input-file (car f)
      (lambda (in)
        (let loop ((neg 0) (pos 0) (zero 0))
          (let ((a (read in)))
            (if (eof-object? a)
                (let ((all (+ neg pos zero)))
                  (set! res (cons (list n 
                                        (percent (/ neg all))
                                        (percent (/ pos all)))
                                  res)))
                (let ((b (read in)))
                  (if (not (number? b))
                      (error "Bogus data file."))
                  (cond ((< a 0)
                         (loop (+ neg 1) pos zero))
                        ((> a 0)
                         (loop neg (+ pos 1) zero))
                        (else
                         (loop neg pos (+ zero 1))))))))))))


; Supporting data: GEN and DOF box/line plot, by collector, of time
; spent in promotion normalized to promotion volume for the run: this
; gives us promotion cost per unit volume.
;
; WARNING: not debugged, because we don't yet have the data!

(define (gen-and-dof-time-spent-in-promotion filename)

  ; data ::= ( ( basic-statistics n ) ... )
  (define (compute-time-spent-in-promotion name data)
    (mappend 
     (lambda (data)
       (call-with-accumulator
        (lambda (acc)
          (for-each 
           (lambda (datum)
             (if (all-iterations-succeeded? datum)
                 (acc (/ (cdr (assq 'run-cpu-gc-promote-best datum))
                         (+ (cdr (assq 'run-bytes-copied-promotion-best datum))
                            0.0)
                         ))))
           (car data)))))
     data))

  (call-with-output-file filename
    (lambda (out)
      (box-and-line-diagrams
       out
       (list ;(compute-time-spent-in-promotion "2GEN" (2gen-canonical-data))
             (compute-time-spent-in-promotion "1DOF" (1dof-canonical-data))
             (compute-time-spent-in-promotion "2DOF" (2dof-canonical-data))
             (compute-time-spent-in-promotion "3DOF" (3dof-canonical-data))
             (compute-time-spent-in-promotion "4DOF" (4dof-canonical-data)))
       #f #f #f #f))))


;----------------------------------------------------------------------
;
; GEN vs ROF

(define (report-gen-and-rof-by-cpu-time file gen-data rof-data)
  (all-benchmarks-gnuplot-report
   file
   (compare-all-by-cpu-time gen-data rof-data)))

(define (report-gen-and-rof-by-cpu-gc-time file gen-data rof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-gc-time gen-data rof-data)))

(define (report-gen-and-rof-by-cpu-mutator-time file gen-data rof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-cpu-mutator-time gen-data rof-data)))

(define (report-gen-and-rof-by-space file gen-data rof-data)
  (all-benchmarks-gnuplot-report 
   file
   (compare-all-by-space gen-data rof-data)))

(define (canned-gen-vs-rof-plots:overall)
  (let ((gen-data  (all-gen-canonical-data))
        (2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (rof-data  (all-rof-canonical-data))
        (dir (lambda xs
               (apply string-append "plots/gen-vs-rof/" xs))))
    
    ; Best of GEN vs ROF (class vs class)

    (report-gen-and-rof-by-cpu-time (dir "cpu-time.data") gen-data rof-data)
    (report-gen-and-rof-by-cpu-gc-time (dir "cpu-gc-time.data")
                                       gen-data rof-data)
    (report-gen-and-rof-by-cpu-mutator-time (dir "cpu-mutator-time.data")
                                            gen-data rof-data)
    (report-gen-and-rof-by-space (dir "space.data") gen-data rof-data)
  
    ; Best of mGEN vs ROF across all benchmarks

    (do ((gencnt '(2 3)                     (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gen))
      (let ((name (string-append (number->string (car gencnt))
                                   "gen-rof.data")))
        (report-gen-and-rof-by-cpu-time 
         (dir "cpu-time-" name) (car gen) rof-data)
        (report-gen-and-rof-by-cpu-gc-time
         (dir "cpu-gc-time-" name) (car gen) rof-data)
        (report-gen-and-rof-by-cpu-mutator-time 
         (dir "cpu-mutator-time-" name) (car gen) rof-data)
        (report-gen-and-rof-by-space 
         (dir "space-" name) (car gen) rof-data)))

    #t))

(define (canned-gen-vs-rof-plots:by-benchmark)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (rof-data  (all-rof-canonical-data))
        (dir (lambda xs (apply string-append "plots/gen-vs-rof/" xs))))

    ; Best of mGEN vs best of ROF by benchmark

    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gen))
      (do ((cmp (list compare-benchmarks-by-cpu-time
                      compare-benchmarks-by-cpu-gc-time
                      compare-benchmarks-by-cpu-mutator-time
                      compare-benchmarks-by-space)
                (cdr cmp))
           (name '("cpu-time" "cpu-gc-time" "cpu-mutator-time" "space")
                 (cdr name)))
          ((null? cmp))
        (let ((name (string-append "bm-"
                                   (car name)
                                   "-"
                                   (number->string (car gencnt))
                                   "gen-rof.tex")))
          (display (dir name))
          (display "...")
          (newline)
          (report-box-and-line-by-benchmark
           (car cmp)
           (dir name)
           (car gen)
           rof-data))))

    #t))

(define (gen-vs-rof-space-time-plots prefix)
  (do ((gencnt '(2 3) (cdr gencnt))
       (gen    (list (2gen-canonical-data) (3gen-canonical-data))
               (cdr gen)))
      ((null? gen))
    (let ((rof (rof-canonical-data)))
      (let ((name (string-append prefix
                                 (number->string (car gencnt))
                                 "gen-rof-")))
        (display (string-append name "..."))
        (newline)
        (report-by-benchmark-space-time name (car gen) rof))))
  #t)

(define (canned-gen-vs-rof-plots)
  (display "GEN vs ROF plots")
  (newline)

  (canned-gen-vs-rof-plots:overall)
  (canned-gen-vs-rof-plots:by-benchmark)
  (gen-vs-rof-space-time-plots "plots/gen-vs-rof/st-")
  #t)

(define (gen-vs-rof-qualitative subprogram)
  (load-gen-results)
  (load-rof-results)
  (let ((2gen-data (2gen-canonical-data))
        (3gen-data (3gen-canonical-data))
        (rof-data (rof-canonical-data)))
    
    (map
     (lambda (gen)
       (map
        (lambda (rof)
          (subprogram gen rof))
        (list rof-data)))
     (list 2gen-data 3gen-data))))

(define (report-gen-vs-rof-qualitative-by-benchmark)

  (define (process out)
    (let ((res (gen-vs-rof-qualitative 
                qualitative-test-by-benchmark)))
      (format out "\\begin{tabular}{|ll|rr|}~%\\hline~%")
      (format out " && \\multicolumn{2}{c|}{ROF} \\\\~%")
      (format out " && GEN & ROF \\\\~%")
      (do ((i 2 (+ i 1))
           (res res (cdr res))
           (name #f))
          ((> i 3))
        (format out "\\hline~%")
        (set! name (format #f "~aGEN" i))
        (for-each 
         (lambda (bm)
           (format out " ~a & ~20a" name (format-benchmark-name bm))
           (set! name "    ")
           (let* ((r (list-ref (car res) 0))
                  (v (cadr (assoc bm r)))
                  (n (vector-ref v 5)))
             (format out " & ~4,1f & ~4,1f" 
                     (* 100 (/ (+ (vector-ref v 0) (vector-ref v 1)) n))
                     (* 100 (/ (+ (vector-ref v 2) (vector-ref v 3)) n))))
           (format out "\\\\~%"))
         benchmark-names))
      (format out "\\hline~%")
      (format out "\\end{tabular}~%")))
      
  (process (current-output-port)))

(define (gen-vs-rof-count-wins-table)

  (define (display-results data out)
    (latex-table
     out
     "|l|rrrr|"
     '(("" "Total time" "GC time" "Heap space" "Total space"))
     `((benchmark-name "~20a") ,@(make-list 4 '(literal "~3a")))
     data))

  (define (tally-victories res bm)
    (let ((bm-results (cdr (assoc bm res))))
      (inexact->exact
       (round 
        (* 100.0 (/ (sum (lambda (x) (if (>= (car x) 0) (cdr x) 0)) 
                         bm-results)
                    (sum (lambda (x) (cdr x)) bm-results)))))))

  (load-rof-results)
  (load-gen-results)
  (let ((2gen-data (all-2gen-canonical-data))
        (3gen-data (all-3gen-canonical-data))
        (rof-data (all-rof-canonical-data))
        (out (current-output-port)))
    (do ((gencnt '(2 3) (cdr gencnt))
         (gen    (list 2gen-data 3gen-data) (cdr gen)))
        ((null? gencnt))
      (let* ((gen-data (car gen))
             (res-t (compare-benchmarks-by-cpu-time gen-data rof-data))
             (res-gc (compare-benchmarks-by-cpu-gc-time gen-data rof-data))
             (res-space (compare-benchmarks-by-space gen-data rof-data))
             (res-tspace (compare-benchmarks-by-total-space gen-data rof-data)))
        (display-results
         (map
          (lambda (bm)
            (list bm
                  (tally-victories res-t bm)
                  (tally-victories res-gc bm)
                  (tally-victories res-space bm)
                  (tally-victories res-tspace bm)))
          benchmark-names)
         out)))))

; Table of promotion and collection cost per volume promoted.
;
; Since the volume promoted varies with the load factor, it may be
; useful to look at whether the promotion cost varies with the load.  We
; get a table with one row for each benchmark and one column for
; promotion cost for each of four loads 2.25, 2.5, 2.75, and 3.0.  In
; addition the leftmost column is promotion cost for 2GEN at L=3.0.

(define (gen-vs-rof-copy-cost-report . rest)

  (define (doit out)
    (let* ((gen (sort-by-benchmark-name (gen-promotion/collection-cost) car))
           (rofs
            (map (lambda (l)
                   (sort-by-benchmark-name 
                    (rof-promotion/collection-cost l) 
                    car))
                 '(2.25 2.5 2.75 3.0)))
           (rofavg 
            (apply map (lambda xs
                         (list (caar xs)
                               (mean (map cadr xs))
                               (mean (map caddr xs))))
                   rofs)))
      (latex-table
       out
       "|l|rrr|"
       '(("" "2GEN" ("ROF" 5) "\\underline{ROF}")
         ("" ""     "$L=2.25$" "$L=2.5$" "$L=2.75$" "$L=3.0$" "Mean" "GEN"))
       `((benchmark-name "~20a")
         ,@(make-list 6 '(literal "~4,1f"))
         (,(lambda (row-datum column)
             (/ (list-ref row-datum 6) (cadr row-datum)))
          "~4,2f"))
       (apply map (lambda rs
                    (cons (caar rs) (map cadr rs)))
              (append (list gen) rofs (list rofavg))))))

  (if (null? rest)
      (doit (current-output-port))
      (call-with-output-medium (car rest) 
        doit)))

; eof
