; Data processing for generational GC.   -*- mode: scheme; mode: font-lock -*-
; 2000-09-11 / lth

;----------------------------------------------------------------------

(define (canned-gen-tables-and-plots)
  (display "GEN tables and plots")
  (newline)

  (load-gen-results)

  ; These create HTML tables
  (create-gen-summary-tables 'html)
  (merge-gen-summary-html-tables)
  (create-gen-gc-tables 'html)
  (merge-gen-gc-html-tables)

  ; These create LaTeX source
  (create-gen-summary-tables 'latex)
  (create-gen-gc-tables 'latex)

  ; These create data files for GNUplot
  ;(create-2gen-gctime-vs-mark/cons-report ...) ; Not used?
  ;(create-3gen-gctime-vs-mark/cons-report ...) ; Not used?
  (create-2gen-gctime-vs-mark/cons-plot "plots/gen-summary-plots/2mc-")
  (create-3gen-gctime-vs-mark/cons-plot "plots/gen-summary-plots/3mc-")
  (create-2gen-load-vs-mark/cons-plot "plots/gen-summary-plots/2L-")
  (create-3gen-load-vs-mark/cons-plot "plots/gen-summary-plots/3L-")

  'canned-gen-tables-and-plots)

;----------------------------------------------------------------------

(define (crunch-gen-results)
  (for-each 
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))

   `(("test/output/2000-04-28-2gen.upper+lower.1" 
      "crunched/2gen.upper+lower.1")
     ("test/output/2000-04-28-2gen.upper+lower.5" 
      "crunched/2gen.upper+lower.5")
     ("test/output/2000-04-07-2gen.upper.1" "crunched/2gen.upper.1")
     ("test/output/2000-04-18-2gen.upper.5" "crunched/2gen.upper.5")
     ("test/output/2000-04-07-2gen.var.1" "crunched/2gen.var.1")
     ("test/output/2000-04-07-2gen.var.5" "crunched/2gen.var.5")
     ("test/output/2000-04-28-3gen.upper+lower.1" 
      "crunched/3gen.upper+lower.1")
     ("test/output/2000-04-28-3gen.upper+lower.5" 
      "crunched/3gen.upper+lower.5")
     ("test/output/2000-04-07-3gen.upper.1" "crunched/3gen.upper.1")
     ("test/output/2000-04-18-3gen.upper.5" "crunched/3gen.upper.5")
     ("test/output/2000-04-07-3gen.var.1" "crunched/3gen.var.1")
     ("test/output/2000-04-07-3gen.var.5" "crunched/3gen.var.5"))))

(define (crunch-gen-promotion-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-06-22-promvol-2gen.var.1" 
      "crunched/promvol-2gen.var.1")
     ("test/output/2000-06-22-promvol-2gen.var.5" 
      "crunched/promvol-2gen.var.5"))))

(define (crunch-gen-float-results)
  (for-each
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))
   '(("test/output/2000-07-11-float-2gen.upper+lower.1" 
      "crunched/float-2gen.upper+lower.1")
     ("test/output/2000-07-11-float-2gen.upper+lower.5" 
      "crunched/float-2gen.upper+lower.5")
     ("test/output/2000-07-11-float-3gen.upper+lower.1" 
      "crunched/float-3gen.upper+lower.1")
     ("test/output/2000-07-11-float-3gen.upper+lower.5" 
      "crunched/float-3gen.upper+lower.5"))))

; Data gathered August 2000 and later.

(define (crunch-gen-extra-results)
  (for-each 
   (lambda (fs)
     (let ((input (car fs))
           (output (cadr fs)))
       (display "*** Processing ")
       (display input)
       (newline)
       (let ((x (crunch input)))
         (call-with-output-file output
           (lambda (out)
             (write x out))))))

   `(; Standard data for EARLEY
     ("test/output/2000-08-21-earley-2gen.upper.5"
      "crunched/2gen-extra.upper.5")
     ("test/output/2000-08-21-earley-2gen.upper+lower.5"
      "crunched/2gen-extra.upper+lower.5")
     ("test/output/2000-08-21-earley-3gen.upper.5"
      "crunched/3gen-extra.upper.5")
     ("test/output/2000-08-21-earley-3gen.upper+lower.5"
      "crunched/3gen-extra.upper+lower.5")

     ; Standard data for PERM and GCOLD (note: GCOLD with broken RNG)
     ("test/output/2000-08-26-2gen.upper.1"
      "crunched/2gen-extra2.upper.1")
     ("test/output/2000-08-26-2gen.upper+lower.1"
      "crunched/2gen-extra2.upper+lower.1")
     ("test/output/2000-08-26-2gen.var.1"
      "crunched/2gen-extra2.var.1")
     ("test/output/2000-08-26-3gen.upper.1"
      "crunched/3gen-extra2.upper.1")
     ("test/output/2000-08-26-3gen.upper+lower.1"
      "crunched/3gen-extra2.upper+lower.1")
     ("test/output/2000-08-26-3gen.var.1"
      "crunched/3gen-extra2.var.1")
     
     ; Float reduction data for EARLEY
     ("test/output/2000-08-23-float-2gen.upper+lower.5"
      "crunched/float-2gen-extra.upper+lower.5")
     ("test/output/2000-08-23-float-3gen.upper+lower.5"
      "crunched/float-3gen-extra.upper+lower.5")

     ; Float reduction data for PERM and GCOLD.
     ("test/output/2000-08-29-float-2gen.upper+lower.1"
      "crunched/float-2gen-extra.upper+lower.1")
     ("test/output/2000-08-29-float-3gen.upper+lower.1"
      "crunched/float-3gen-extra.upper+lower.1")

     ; FIXME: more promotion data here when PERM and GCOLD are done.
     ("test/output/2000-09-04-promvol-2gen.var.1"
      "crunched/promvol-2gen-extra.var.1")

     )))

;----------------------------------------------------------------------

(define gen-results-loaded #f)

(define (load-gen-results)
  (if (not gen-results-loaded)
      (begin
        (set! gen-results-loaded #t)
        (for-each (lambda (x)
                    (let ((file (car x))
                          (variable (cadr x)))
                      (environment-set! (interaction-environment)
                                        variable
                                        (load-crunched-data file))))
                  '(("crunched/2gen.upper.1" 2gen-upper-1)
                    ("crunched/2gen.upper.5" 2gen-upper-5)
                    ("crunched/2gen.upper+lower.1" 2gen-upper+lower-1)
                    ("crunched/2gen.upper+lower.5" 2gen-upper+lower-5)
                    ("crunched/2gen.var.1" 2gen-var-1)
                    ("crunched/2gen.var.5" 2gen-var-5)
                    ("crunched/3gen.upper.1" 3gen-upper-1)
                    ("crunched/3gen.upper.5" 3gen-upper-5)
                    ("crunched/3gen.upper+lower.1" 3gen-upper+lower-1)
                    ("crunched/3gen.upper+lower.5" 3gen-upper+lower-5)
                    ("crunched/3gen.var.1" 3gen-var-1)
                    ("crunched/3gen.var.5" 3gen-var-5)))

        ; Merge in results of later EARLEY runs.
        (set! 2gen-upper-5 
              (append (remove-benchmark '(5 earley:12) 2gen-upper-5)
                      (load-crunched-data "crunched/2gen-extra.upper.5")))
        (set! 2gen-upper+lower-5
              (append (remove-benchmark '(5 earley:12) 2gen-upper+lower-5)
                      (load-crunched-data 
                       "crunched/2gen-extra.upper+lower.5")))

        (set! 3gen-upper-5 
              (append (remove-benchmark '(5 earley:12) 3gen-upper-5)
                      (load-crunched-data "crunched/3gen-extra.upper.5")))
        (set! 3gen-upper+lower-5
              (append (remove-benchmark '(5 earley:12) 3gen-upper+lower-5)
                      (load-crunched-data 
                       "crunched/3gen-extra.upper+lower.5")))

        ; Merge in results of later PERM and GCOLD runs.
        (set! 2gen-upper-1 
              (append 2gen-upper-1
                      (load-crunched-data "crunched/2gen-extra2.upper.1")))
        (set! 2gen-upper+lower-1
              (append 2gen-upper+lower-1
                      (load-crunched-data 
                       "crunched/2gen-extra2.upper+lower.1")))
        (set! 2gen-var-1 
              (append 2gen-var-1
                      (load-crunched-data "crunched/2gen-extra2.var.1")))

        (set! 3gen-upper-1 
              (append 3gen-upper-1
                      (load-crunched-data "crunched/3gen-extra2.upper.1")))
        (set! 3gen-upper+lower-1
              (append 3gen-upper+lower-1
                      (load-crunched-data 
                       "crunched/3gen-extra2.upper+lower.1")))
        (set! 3gen-var-1 
              (append 3gen-var-1
                      (load-crunched-data "crunched/3gen-extra2.var.1")))

        #t)))

(define gen-promvol-results-loaded #f)

(define (load-gen-promvol-results)
  (if (not gen-promvol-results-loaded)
      (begin
        (for-each 
         (lambda (x)
           (let ((file (car x))
                 (variable (cadr x)))
             (environment-set! (interaction-environment)
                               variable
                               (load-crunched-data file))))
         '(("crunched/promvol-2gen.var.1" promvol-2gen-var-1)
           ("crunched/promvol-2gen.var.5" promvol-2gen-var-5)))

        (set! promvol-2gen-var-1
              (append promvol-2gen-var-1
                      (load-crunched-data "crunched/promvol-2gen-extra.var.1")))
        (set! gen-promvol-results-loaded #t)
        (unspecified))))


(define gen-float-results-loaded #f)

(define (load-gen-float-results)
  (if (not gen-float-results-loaded)
      (begin
        (set! float-2gen-upper+lower-1 
              (crunched-add-option!
               (load-crunched-data "crunched/float-2gen.upper+lower.1")
               '-gen-fullgc-on-promotion))
        (set! float-2gen-upper+lower-5
              (crunched-add-option!
               (load-crunched-data "crunched/float-2gen.upper+lower.5")
               '-gen-fullgc-on-promotion))
        (set! float-3gen-upper+lower-1 
              (crunched-add-option!
               (load-crunched-data "crunched/float-3gen.upper+lower.1")
               '-gen-fullgc-on-promotion))
        (set! float-3gen-upper+lower-5 
              (crunched-add-option!
               (load-crunched-data "crunched/float-3gen.upper+lower.5")
               '-gen-fullgc-on-promotion))

        ; Merge in EARLEY float results.

        (set! float-2gen-upper+lower-5
              (append (remove-benchmark '(5 earley:12) 
                                        float-2gen-upper+lower-5)
                      (load-crunched-data 
                       "crunched/float-2gen-extra.upper+lower.5")))

        (set! float-3gen-upper+lower-5
              (append (remove-benchmark '(5 earley:12) 
                                        float-3gen-upper+lower-5)
                      (load-crunched-data 
                       "crunched/float-3gen-extra.upper+lower.5")))

        ; Merge in in PERM and GCOLD float results.

        (set! float-2gen-upper+lower-1
              (append float-2gen-upper+lower-1
                      (load-crunched-data
                       "crunched/float-2gen-extra.upper+lower.1")))
        (set! float-3gen-upper+lower-1
              (append float-3gen-upper+lower-1
                      (load-crunched-data
                       "crunched/float-3gen-extra.upper+lower.1")))

        ; FIXME: more here when PERM and GCOLD are done.

        (set! gen-float-results-loaded #t)
        (unspecified))))


; ORDER MATTERS!

(define (2gen-canonical-data)
  (load-gen-results)
  (fluid-let ((no-warnings #t))
    `((,(basic-statistics 2gen-upper-1) 1) 
      (,(basic-statistics 2gen-upper-5) 5)
      (,(basic-statistics 2gen-upper+lower-1) 1)
      (,(basic-statistics 2gen-upper+lower-5) 5)
      (,(basic-statistics 2gen-var-1) 1)
      (,(basic-statistics 2gen-var-5) 5))))

(define (3gen-canonical-data)
  (load-gen-results)
  (fluid-let ((no-warnings #t))
    `((,(basic-statistics 3gen-upper-1) 1) 
      (,(basic-statistics 3gen-upper-5) 5)
      (,(basic-statistics 3gen-upper+lower-1) 1) 
      (,(basic-statistics 3gen-upper+lower-5) 5)
      (,(basic-statistics 3gen-var-1) 1) 
      (,(basic-statistics 3gen-var-5) 5))))

(define (2gen-float-data-reduced)
  (load-gen-float-results)
  `((,(flatten-option-noise (basic-statistics float-2gen-upper+lower-1)) 1)
    (,(flatten-option-noise (basic-statistics float-2gen-upper+lower-5)) 5)))

(define (2gen-float-data-normal)
  (load-gen-results)
  `((,(select-trigger 3.0
       (flatten-option-noise (basic-statistics 2gen-upper+lower-1))) 1)
    (,(select-trigger 3.0
       (flatten-option-noise (basic-statistics 2gen-upper+lower-5))) 5)))

(define (3gen-float-data-reduced)
  (load-gen-float-results)
  `((,(flatten-option-noise (basic-statistics float-3gen-upper+lower-1)) 1)
    (,(flatten-option-noise (basic-statistics float-3gen-upper+lower-5)) 5)))

(define (3gen-float-data-normal)
  (load-gen-results)
  `((,(select-trigger 3.0
       (flatten-option-noise (basic-statistics 3gen-upper+lower-1))) 1)
    (,(select-trigger 3.0
       (flatten-option-noise (basic-statistics 3gen-upper+lower-5))) 5)))

(define (create-gen-summary-tables . rest)

  (define type (if (null? rest) 'latex (car rest)))
  (define generations 0)

  (define (stats breakout? list-of-systems . rest)
    (apply standard-report list-of-systems
           '((("CPU")         ms run-cpu-best)
             (("CPU GC")      ms run-cpu-gc-best)
             (("CPU" "Promote")  ms run-cpu-gc-promote-best)
             (("Major" "GCs") n  run-major-collections-best)
             (("Marked")      mb run-bytes-marked-best)
             (("Heap" "used") mb bytes-max-heap))
           'no-warnings 'break 9 type 'no-variance
           'no-fail
           'filter flatten-option-noise
           'sort sort-by-trigger
           'header-filter (lambda (bm collector heap aux)
                            (values (format-benchmark-name bm)
                                    (format #f "~aGEN" generations)
                                    heap 
                                    (if (null? aux) 
                                        #f
                                        (format #f "T=~a" (car aux)))))
           rest))

  (load-gen-results)
  (for-each (lambda (var)
              (let* ((s    (symbol->string var))
                     (file (case type
                             ((latex) (string-append "tables/" s ".tex"))
                             ((html)  (string-append "html-tables/" s ".html"))
                             (else ???)))
                     (n    (- (char->integer 
                               (string-ref s (- (string-length s) 1)))
                              (char->integer #\0)))
                     (data (environment-get (interaction-environment) 
                                            var))
                     (limit (not (string=? "var" (substring s 5 8))))
                     (3gen? (char=? #\3 (string-ref s 0))))
                (display var)
                (newline)
                (set! generations (if 3gen? 3 2))
                (cond ((null? data)
                       (display "Missing data for ")
                       (display var)
                       (newline))
                      (else
                       (with-output-to-file file
                         (lambda ()
                           (if limit
                               (stats 3gen? data 'no-load/trigger)
                               (stats 3gen? data 'no-heap-limit))))))))
            '(2gen-upper-1 2gen-upper-5
              2gen-upper+lower-1 2gen-upper+lower-5
              2gen-var-1 2gen-var-5
              3gen-upper-1 3gen-upper-5
              3gen-upper+lower-1 3gen-upper+lower-5
              3gen-var-1 3gen-var-5))
  #t)


(define (merge-gen-summary-html-tables)
  (merge-summary-html-tables
   "html-tables/2gen-general.html"
   "2gen results | General"
    '("2gen-upper+lower-1" "2gen-upper+lower-5"
      "2gen-upper-1" "2gen-upper-5"
      "2gen-var-1" "2gen-var-5"))
  (merge-summary-html-tables
   "html-tables/3gen-general.html"
   "3gen results | General"
    '("3gen-upper+lower-1" "3gen-upper+lower-5"
      "3gen-upper-1" "3gen-upper-5"
      "3gen-var-1" "3gen-var-5")))

; Note, remset peak is peak _allocated_, not peak used.

(define (create-gen-gc-tables . rest)

  (define type (if (null? rest) 'latex (car rest)))

  (define (stats1 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms cpu-best)
             (("CPU" "GC")        ms cpu-gc-best)
             (("CPU" "User")      ms cpu-user-best)
             (("CPU" "Sys.")      ms cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get)
                                        (- (get 'cpu-best) 
                                           (get 'cpu-gc-best))))
             (("Copy" "gc")       mb bytes-copied-best)
             (("Move" "los")      mb bytes-moved-best)
             (("Objs" "inserted") exact.kf.1 remset-entries-added-best)
             (("Objs" "scanned")  exact.kf.1 remset-entries-scanned-best)
             (("Bytes" "scanned") exact.kf.1 remset-bytes-scanned-best)
             (("Objs" "removed")  exact.kf.1 remset-entries-removed-best)
             (("Bytes/" "object") nf.1 
                                  ,(lambda (get present?)
                                     (exact->inexact
                                      (/ (get 'remset-bytes-scanned-best)
                                         (get 'remset-entries-scanned-best)))))
             (("Remset" "peak")   kb bytes-max-remset))
           'no-warnings 'break 8 type 'no-variance 'no-load/trigger
           rest))

  (define (stats5 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms run-cpu-best)
             (("CPU" "GC")        ms run-cpu-gc-best)
             (("CPU" "User")      ms run-cpu-user-best)
             (("CPU" "Sys.")      ms run-cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get present?)
                                        (- (get 'run-cpu-best) 
                                           (get 'run-cpu-gc-best))))
             (("Copy" "gc")       mb run-bytes-copied-best)
             (("Move" "los")      mb run-bytes-moved-best)
             (("Objs" "inserted") kf.1 run-remset-entries-added-best)
             (("Objs" "scanned")  kf.1 run-remset-entries-scanned-best)
             (("Bytes" "scanned") kf.1 run-remset-bytes-scanned-best)
             (("Objs" "removed")  kf.1 run-remset-entries-removed-best)
             (("Bytes/" "object") nf.1 
              ,(lambda (get present?)
                 (exact->inexact
                  (/ (get 'run-remset-bytes-scanned-best)
                     (get 'run-remset-entries-scanned-best)))))
             (("Remset" "peak")   kb bytes-max-remset))
           'no-warnings 'break 8 type 'no-variance 'no-load/trigger
           rest))

  (for-each (lambda (var)
              (let* ((s    (symbol->string var))
                     (file (case type
                             ((latex) (string-append "tables/gc-" s ".tex"))
                             ((html)  (string-append "html-tables/gc-" s ".html"))
                             (else ???)))
                     (n    (- (char->integer 
                               (string-ref s (- (string-length s) 1)))
                              (char->integer #\0)))
                     (data (environment-get (interaction-environment) 
                                            var))
                     (limit (not (string=? "var" (substring s 5 8)))))
                (display var)
                (newline)
                (with-output-to-file file
                  (lambda ()
                    (if (= n 1)
                        (if limit
                            (stats1 data)
                            (stats1 data 'no-heap-limit))
                        (if limit
                            (stats5 data)
                            (stats5 data 'no-heap-limit)))))))
            '(2gen-upper-1 2gen-upper-5
              2gen-upper+lower-1 2gen-upper+lower-5
              2gen-var-1 2gen-var-5
              3gen-upper-1 3gen-upper-5
              3gen-upper+lower-1 3gen-upper+lower-5
              3gen-var-1 3gen-var-5))
  #t)

(define (merge-gen-gc-html-tables)
  (merge-summary-html-tables
   "html-tables/2gen-gc.html"
   "2gen results | GC"
    '("gc-2gen-upper+lower-1" "gc-2gen-upper+lower-5"
      "gc-2gen-upper-1" "gc-2gen-upper-5"
      "gc-2gen-var-1" "gc-2gen-var-5"))
  (merge-summary-html-tables
   "html-tables/3gen-gc.html"
   "3gen results | GC"
    '("gc-3gen-upper+lower-1" "gc-3gen-upper+lower-5"
      "gc-3gen-upper-1" "gc-3gen-upper-5"
      "gc-3gen-var-1" "gc-3gen-var-5")))


; Returns a list: (bm copied moved)

(define (gen-extract-promotion/collection-volume)
  (load-gen-promvol-results)
  (let ((r1 (basic-statistics promvol-2gen-var-1))
        (r2 (basic-statistics promvol-2gen-var-5)))
    (extract-promotion/collection-volume (append r1 r2))))

(define (gen-extract-promotion/collection-cost)
  (load-gen-results)
  (let ((r1 (basic-statistics 2gen-var-1))
        (r2 (basic-statistics 2gen-var-5)))
    (map (lambda (x)
           (list (caar x) 
                 (mean (map cadr x))
                 (mean (map caddr x))))
         (accumulate-unordered
          (lambda (a b)
            (equal? (car a) (car b)))
          (filter-map (lambda (x)
                        (if (all-iterations-succeeded? x)
                            (list (benchmark-designator x)
                                  (cdr (assq 'run-cpu-gc-promote-best x))
                                  (cdr (assq 'run-cpu-gc-collect-best x)))
                            #f))
                      (append r1 r2))))))

(define (gen-promotion/collection-cost)
  (let ((v (gen-extract-promotion/collection-volume))
        (c (gen-extract-promotion/collection-cost)))
    (map (lambda (x)
           (let ((y (cdr (assoc (car x) v))))
             (list (car x)
                   (exact->inexact (/ (cadr x) 
                                      (/ (+ (caar y) (cdar y))
                                         (* 1024 1024))))
                   (exact->inexact (/ (caddr x)
                                      (/ (+ (caadr y) (cdadr y))
                                         (* 1024 1024)))))))
         c)))

; Return list of (bencmark-name . cost-in-ms-per-mb)

(define (gen-copy-cost results)
  (map (lambda (x)
         (cons (caar x) 
               (exact->inexact (mean (map cadr x)))))
       (accumulate-unordered
        (lambda (a b)
          (equal? (car a) (car b)))
        (filter-map 
         (lambda (x)
           (if (all-iterations-succeeded? x)
               (list (benchmark-designator x)
                     (/ (cdr (assq 'run-cpu-gc-best x))
                        (/ (+ (cdr (assq 'run-bytes-copied-best x))
                              (cdr (assq 'run-bytes-moved-best x)))
                           (* 1024 1024))))
               #f))
         results))))

(define (2gen-copy-cost)
  (load-gen-results)
  (gen-copy-cost (append (basic-statistics 2gen-var-1)
                         (basic-statistics 2gen-var-5))))

(define (3gen-copy-cost)
  (load-gen-results)
  (gen-copy-cost (append (basic-statistics 3gen-var-1)
                         (basic-statistics 3gen-var-5))))

; Allocation volume

(define (2gen-allocation-volume)
  (filter-map-benchmark-result
   (lambda (datum n)
     (if (all-iterations-succeeded? datum)
         (cons (benchmark-designator datum)
               (cdr (assq 'run-bytes-allocated-best datum)))
         #f))
   (2gen-canonical-data)))

; Mark/cons report

(define (report-gen-vs-mark/cons p1 p5 results)
  (report-vs-mark/cons 
   (lambda (get bm)
     (if (= 1 (car bm))
         (+ (get 'bytes-copied-best)
            (get 'bytes-moved-best)
            (* (static-size bm) (get 'garbage-collections-best)))
         (+ (get 'run-bytes-copied-best)
            (get 'run-bytes-moved-best)
            (* (static-size bm) (get 'run-garbage-collections-best)))))
   p1 p5 results))

(define (create-gen-vs-mark/cons-report fn p1 p5 . rest)
  (let ((raw-results (if (null? rest)
                         (list 2gen-upper-1 2gen-upper-5
                               2gen-upper+lower-1 2gen-upper+lower-5
                               2gen-var-1 2gen-var-5
                               3gen-upper-1 3gen-upper-5
                               3gen-upper+lower-1 3gen-upper+lower-5
                               3gen-var-1 3gen-var-5)
                         (car rest))))
    (create-vs-mark/cons-report 
     fn 
     (lambda (stats)
       (report-gen-vs-mark/cons p1 p5 stats))
     raw-results)))

(define (create-2gen-gctime-vs-mark/cons-report fn)
  (create-gen-vs-mark/cons-report 
   fn
   (lambda (get bm) (/ (get 'cpu-gc-best) 1000.0))
   (lambda (get bm) (/ (get 'run-cpu-gc-best) 1000.0))
   (list 2gen-upper-1 2gen-upper-5
         2gen-upper+lower-1 2gen-upper+lower-5
         2gen-var-1 2gen-var-5)))

(define (create-3gen-gctime-vs-mark/cons-report fn)
  (create-gen-vs-mark/cons-report 
   fn
   (lambda (get bm) (/ (get 'cpu-gc-best) 1000.0))
   (lambda (get bm) (/ (get 'run-cpu-gc-best) 1000.0))
   (list 3gen-upper-1 3gen-upper-5
         3gen-upper+lower-1 3gen-upper+lower-5
         3gen-var-1 3gen-var-5)))

(define (gen-gctime get bm)        ; in seconds
  (if (= 1 (car bm))
      (/ (get 'cpu-gc-best) 1000.0)
      (/ (get 'run-cpu-gc-best) 1000.0)))

(define (gen-gctime-normalized get bm)
  (/ (* (gen-gctime get bm) 1000.0)
     (/ (canonical-allocation bm) (* 1024 1024))))

(define (gen-mark-volume get bm)
  (if (= 1 (car bm))
      (+ (get 'bytes-copied-best)
         (get 'bytes-moved-best))
      (+ (get 'run-bytes-copied-best)
         (get 'run-bytes-moved-best))))

(define (gen-mark-cons-ratio get bm)
  (/ (gen-mark-volume get bm) (canonical-allocation bm)))

(define (gen-effective-load get bm) ; That is _load_, as in 1/L.
  (/ (profiled-peak-usage bm) (get 'bytes-max-heap)))

; FIXME: does not include heap.
(define (gen-effective-inverse-load get bm) ; That is L
  (/ (get 'bytes-max-heap) (profiled-peak-usage bm)))

(define (create-gen-benchmark-2d-plot fn independent dependent . rest)
  (let ((raw-results (if (null? rest)
                         (list 2gen-upper-1 2gen-upper-5
                               2gen-upper+lower-1 2gen-upper+lower-5
                               2gen-var-1 2gen-var-5
                               3gen-upper-1 3gen-upper-5
                               3gen-upper+lower-1 3gen-upper+lower-5
                               3gen-var-1 3gen-var-5)
                         (car rest))))
    (create-benchmark-2d-plot
     fn 
     (lambda (results)
       (extract-data results independent dependent))
     raw-results)))

(define (create-2gen-gctime-vs-mark/cons-plot prefix)
  (create-gen-benchmark-2d-plot prefix
                                gen-mark-cons-ratio
                                gen-gctime
                                (list 2gen-upper-1 2gen-upper-5
                                      2gen-upper+lower-1 2gen-upper+lower-5
                                      2gen-var-1 2gen-var-5)))

(define (create-3gen-gctime-vs-mark/cons-plot prefix)
  (create-gen-benchmark-2d-plot prefix
                                gen-mark-cons-ratio
                                gen-gctime
                                (list 3gen-upper-1 3gen-upper-5
                                      3gen-upper+lower-1 3gen-upper+lower-5
                                      3gen-var-1 3gen-var-5)))

(define (create-2gen-load-vs-mark/cons-plot prefix)
  (create-gen-benchmark-2d-plot prefix
                                gen-effective-inverse-load
                                gen-mark-cons-ratio
                                (list 2gen-upper-1 2gen-upper-5
                                      2gen-var-1 2gen-var-5)))

(define (create-2gen-load-vs-gctime-plot prefix)
  (load-gen-results)
  (create-gen-benchmark-2d-plot prefix
                                gen-effective-inverse-load
                                gen-gctime-normalized
                                (list 2gen-upper-1 2gen-upper-5
                                      2gen-var-1 2gen-var-5)))

  
(define (create-3gen-load-vs-mark/cons-plot prefix)
  (create-gen-benchmark-2d-plot prefix
                                gen-effective-inverse-load
                                gen-mark-cons-ratio
                                (list 3gen-upper-1 3gen-upper-5
                                      3gen-var-1 3gen-var-5)))

(define (create-gen-heap-vs-mark/cons-report fn)
  (let ((c (lambda (get bm) 
             (/ (- (/ (get 'bytes-max-heap) (profiled-peak-usage bm)) 0.0)))))
    (create-gen-vs-mark/cons-report fn c c (list 2gen-var-1 3gen-var-1))))

(define (report-generational-copy-speed)
  (let ((data (sort (mappend report-copy-speed (append (2gen-canonical-data)
                                                       (3gen-canonical-data)))
                    <)))
    `((min ,(car data))
      (max ,(car (last-pair data))) 
      (mean ,(mean data))
      (median ,(median data))
      (std-dev ,(standard-deviation data)))))


; Major-gc vs minor-gc report

(define (gen-major-vs-minor-report)
  (latex-table
   (current-output-port)
   "|l|rrr|"
   '(("" "Major" "Minor" "Ratio"))
   '((benchmark-name "~20a") 
     (literal "~6a") (literal "~6a") (literal "~6,2f"))
   (let ((results
          (accumulate-unordered
           (lambda (a b)
             (equal? (car a) (car b)))
           (filter-map-benchmark-result
            (lambda (datum n)
              (if (all-iterations-succeeded? datum)
                  (list (benchmark-designator datum)
                        (cdr (assq 'run-cpu-gc-major-best datum))
                        (cdr (assq 'run-cpu-gc-promote-best datum)))
                  #f))
            (2gen-canonical-data)))))
     (map (lambda (bm)
            (let ((probe (find (lambda (x)
                                 (equal? (caar x) bm))
                               results)))
              (if probe
                  (let ((x (mean (map cadr probe)))
                        (y (mean (map caddr probe))))
                    (list bm 
                          (inexact->exact (round x))
                          (inexact->exact (round y))
                          (/ x y)))
                  (list bm #f #f #f))))
          benchmark-names))))

; Float analysis

(define (gen-float-report)
  (let ((2gen-data (gen-float-analysis-core (2gen-float-data-normal)
                                            (2gen-float-data-reduced)))
        (3gen-data (gen-float-analysis-core (3gen-float-data-normal)
                                            (3gen-float-data-reduced))))
    (latex-table
     (current-output-port)
     "|l|rr|rr|"
     '(("" ("Percent" 2) ("Volume" 2))
       ("" "2GEN" "3GEN" "2GEN" "3GEN"))
     (cons '(benchmark-name "~20a")
           (make-list 4 '(literal "~6,1f")))
     (map (lambda (bm)
            (define (get data fn)
              (let ((probe (assoc bm data)))
                (if probe (fn probe) 0.0)))
            (let ((2normal (get 2gen-data cadr))
                  (2reduced (get 2gen-data caddr))
                  (3normal (get 3gen-data cadr))
                  (3reduced (get 3gen-data caddr)))
              (list bm
                    (* 100 (/ (- 2normal 2reduced) 2reduced))
                    (* 100 (/ (- 3normal 3reduced) 3reduced))
                    (/ (- 2normal 2reduced) (* 1024 1024))
                    (/ (- 3normal 3reduced) (* 1024 1024)))))
          benchmark-names))))

; This is probably general.

(define (gen-float-analysis-core normal-data reduced-data)
  (map (lambda (x)
         (list (caar x) 
               (exact->inexact (mean (map cadr x)))
               (exact->inexact (mean (map caddr x)))))
       (accumulate-unordered
        (lambda (a b)
          (equal? (car a) (car b)))
        (call-with-accumulator
         (lambda (acc)
           (for-each-benchmark-result*
            (lambda (n a b)
              (if (and (all-iterations-succeeded? a)
                       (all-iterations-succeeded? b))
                  (acc (list (benchmark-designator a)
                             (+ (cdr (assq 'run-bytes-copied-best a))
                                (cdr (assq 'run-bytes-moved-best a)))
                             (+ (cdr (assq 'run-bytes-copied-best b))
                                (cdr (assq 'run-bytes-moved-best b)))))))
            normal-data
            reduced-data))))))


; All-data as from *gen-float-data.

; This is wrong, because it must only use runs where both systems had a run
; that succeeded.  Otherwise the results may be garbage.

'(define (gen-float-analysis-core all-data)

  (define (datum-tag datum)
    (cond ((let ((x (assq '-gen-fullgc-on-promotion datum)))
             (and x (cdr x)))
           'promote)
          (else 
           'normal)))

  (define results #f)

  (define (process-datum datum n)
    (if (all-iterations-succeeded? datum)
        (let ((cell ((results 'lookup) (cons (benchmark-designator datum) 
                                             (datum-tag datum)))))
          (set-car! (cdr cell) (+ (cadr cell) 
                                  (cdr (assq 'run-bytes-copied-best datum))
                                  (cdr (assq 'run-bytes-moved-best datum))))
          (set-car! (cddr cell) (+ (caddr cell) 1)))))

  (set! results (make-benchmark-associator (lambda () (list 0 0))))
  (for-each-benchmark-result process-datum all-data)
  (map (lambda (x) 
         (if (zero? (caddr x))
             (cons (car x) #f)
             (cons (car x) (exact->inexact (/ (cadr x) (caddr x))))))
       ((results 'results))))

; eof
