; Some code for drawing pictures with latex
; 2000-08-01 / lth

(require 'format)
(require 'list)

(define units-across 1500)              ; gnuplot
(define units-down 900)                 ; gnuplot
(define unit-length 0.240900)           ; 0.240900 points per dot gives 300dpi


; data ::= ((name low high boxlow boxhigh devlow devhigh avg) ...)

(define (box-and-line-diagrams output data xmin xmax xticks tickstart . rest)

  ; There's a general principle at work here but I haven't quite nailed
  ; it down yet.
  (define (select-ticks range)
    (cond ((> range 200) 50)
          ((> range 100) 20)
          ((> range 50)  10)
          ((> range 20)  5)
          ((> range 10)  2)
          ((> range 5)   1)
          (else 0.1)))

  (define (round-up v multiple)
    (cond ((integer? multiple)
           (* multiple (quotient (+ (ceiling v) (- multiple 1)) multiple)))
          ((integer? (* 10.0 multiple))
           (/ (round-up (* 10.0 v) (* 10.0 multiple)) 10.0))
          (else
           (error "round-up: domain error " v " " multiple))))

  (define (round-down v multiple)
    (cond ((integer? multiple)
           (round-up (- (floor v) (- multiple 1)) multiple))
          ((integer? (* 10.0 multiple))
           (/ (round-down (* 10.0 v) (* 10.0 multiple)) 10.0))
          (else
           (error "round-down: domain error " v " " multiple))))

  (define (doit)
    (let* ((smallest (apply min (map cadr data)))
           (largest  (apply max (map caddr data)))
           (numdata  (length data))
           (xmin     (if xmin xmin (- smallest 1.0)))
           (xmax     (if xmax xmax (+ largest 1.0)))
           (ymin     0.0)
           (height   (+ numdata 1.0))
           (xunit    (/ units-across (- xmax xmin)))
           (textgap  -30)
           (yunit    (/ units-down height))
           (xticks   (if xticks xticks (select-ticks (- xmax xmin))))
           (tickstart (if tickstart tickstart (round-up xmin xticks))))

      (define (x-logical x) (+ (/ x xunit) xmin))
      (define (y-logical y) (+ (/ y yunit) ymin))
      (define (x-actual x) (* xunit (- x xmin)))
      (define (y-actual y) (* yunit (- y ymin)))

      (define (threeplaces x)
        (if (and (number? x) (not (integer? x)))
            (/ (round (* 1000.0 x)) 1000.0)
            x))

      (define (emit fmt . args)
        (display (apply format #f fmt (map threeplaces args)) output)
        (newline output))

      (define (line x1 y1 x2 y2)
        (cond ((= x1 x2)
               (emit "\\put(~a,~a){\\rule[-0.200pt]{0.400pt}{~apt}}"
                     x1 (min y1 y2) (* unit-length (max (- y2 y1) (- y1 y2)))))
              ((= y1 y2)
               (emit "\\put(~a,~a){\\rule[-0.200pt]{~apt}{0.400pt}}"
                     (min x1 x2) y1 (* unit-length (max (- x2 x1) (- x1 x2)))))
              (else
               (error "Only vertical or horizontal lines."))))

      (define (box x1 y1 x2 y2)
        (cond ((> x1 x2)
               (box x2 y1 x1 y2))
              ((> y1 y2)
               (box x1 y2 x2 y1))
              (else
               (line x1 y1 x2 y1)
               (line x2 y1 x2 y2)
               (line x2 y2 x1 y2)
               (line x1 y2 x1 y1))))

      (define (text-extending-left x1 y1 t)
        (emit "\\put(~a,~a){\\makebox(0,0)[r]{~a}}" x1 y1 t))

      (define (text-extending-down x1 y1 t)
        (emit "\\put(~a,~a){\\makebox(0,0)[t]{~a}}" x1 y1 t))

      (define (text-extending-right x1 y1 t)
        (emit "\\put(~a,~a){\\makebox(0,0)[l]{~a}}" x1 y1 t))

      (define (show-datum i name low high boxlow boxhigh devlow devhigh avg)
        (let ((boxwidth/2 (y-logical (/ units-down (* numdata 8.0)))))
          (text-extending-left textgap (y-actual i) 
                               (format #f "{\\tt ~a}" name))
          (line (x-actual low) (y-actual i)
                (x-actual high) (y-actual i))
          (box  (x-actual boxlow) (y-actual (- i boxwidth/2))
                (x-actual boxhigh) (y-actual (+ i boxwidth/2)))
          (line (x-actual avg) (y-actual (- i (* boxwidth/2 3.0)))
                (x-actual avg) (y-actual (+ i (* boxwidth/2 3.0))))
          ;(box  (x-actual devlow) (y-actual (- i (* boxwidth/2 2/3)))
          ;      (x-actual devhigh) (y-actual (+ i (* boxwidth/2 2/3))))
          ))

      (define (draw-frame)

        (define (draw-tick i)
          (line (x-actual i) -10 (x-actual i) 10)
          (text-extending-down (x-actual i)
                               -20
                               (if (integer? i)
                                   (number->string (inexact->exact i))
                                   (number->string i))))

        (box 0 0 units-across units-down)
        (do ((i tickstart (+ i xticks)))
            ((> i xmax))
          (draw-tick i))
        (do ((i (- tickstart xticks) (- i xticks)))
            ((< i xmin))
          (draw-tick i)))

      ; 0.240900 points per dot gives 300dpi

      (emit "\\setlength{\\unitlength}{0.240900pt}")
      (emit "\\begin{picture}(~a,~a)(0,0)" units-across units-down)
      (emit "\\font\\gnuplot=cmr10 at 10pt")
      (emit "\\gnuplot")
      (draw-frame)
      (for-each (lambda (datum i)
                  (apply show-datum i datum))
                (reverse data) (iota1 numdata))
      (emit "\\end{picture}")))

  (let loop ((rest rest) (across units-across) (down units-down))
    (cond ((null? rest)
           (fluid-let ((units-across across)
                       (units-down down))
             (doit)))
          ((eq? (car rest) 'across)
           (loop (cddr rest) (cadr rest) down))
          ((eq? (car rest) 'down)
           (loop (cddr rest) across (cadr rest)))
          (else ???))))


(define (make-latex-plot)
  (let ((l (list 'head)))
    (lambda (msg)
      (case msg
        ((line) 
         (lambda args (append! l (list (cons 'line args)))))
        ((text-left) 
         (lambda args (append! l (list (cons 'text-left args)))))
        ((text-right)
         (lambda args (append! l (list (cons 'text-right args)))))
        ((*get)
         (lambda () (cdr l)))
        (else ???)))))

; picture is picx units wide and picy drawing units tall
; drawing is expressed in units that are 'unit' drawing units.
; minimum, maximum x and y logical units are as given.

(define (latex-plot output                          ; a port
                    units-across units-down         ; physical
                    log-unit xmin ymax xmax ymin    ; logical
                    plot)

  (define (x-actual x) (* log-unit x))
  (define (y-actual y) (* log-unit y))

  ; Shortest line about 1/7 inch (according to manual)

  (define shortest-line (* 1/7 300.0))  ; in dots

  (define (threeplaces x)
    (if (and (number? x) (not (integer? x)))
        (/ (round (* 1000.0 x)) 1000.0)
        x))

  (define (emit fmt . args)
    (apply format output fmt (map threeplaces args))
    (newline output))

  (define (preamble)                    ; 0.240900 points per dot gives 300dpi
    (emit "\\setlength{\\unitlength}{0.240900pt}")
    (emit "\\begin{picture}(~a,~a)(~a,~a)" 
          units-across units-down
          (* log-unit xmin) (* log-unit ymin))
    (emit "\\font\\gnuplot=cmr10 at 10pt")
    (emit "\\gnuplot"))

  (define (postamble)
    (emit "\\end{picture}"))

  (define (approx= a b)
    (< 0.99 (/ a b) 1.01))

  (define (text-extending-left x1 y1 t)
    (emit "\\put(~a,~a){\\makebox(0,0)[r]{~a}}" 
          (x-actual x1) (y-actual y1) t))

  (define (text-extending-down x1 y1 t)
    (emit "\\put(~a,~a){\\makebox(0,0)[t]{~a}}" 
          (x-actual x1) (y-actual y1) t))

  (define (text-extending-right x1 y1 t)
    (emit "\\put(~a,~a){\\makebox(0,0)[l]{~a}}" 
          (x-actual x1) (y-actual y1) t))

  (define (line x1 y1 x2 y2)
    (cond ((approx= x1 x2)
           (emit "\\put(~a,~a){\\rule[-0.200pt]{0.400pt}{~apt}}"
                 (x-actual x1)
                 (y-actual (min y1 y2))
                 (* unit-length log-unit (max (- y2 y1) (- y1 y2)))))
          ((approx= y1 y2)
           (emit "\\put(~a,~a){\\rule[-0.200pt]{~apt}{0.400pt}}"
                 (x-actual (min x1 x2))
                 (y-actual y1)
                 (* unit-length log-unit (max (- x2 x1) (- x1 x2)))))
          ((approx= (abs (- x2 x1)) (abs (- y2 y1)))
           (let* ((up (inexact->exact (round (- y2 y1))))
                  (across (inexact->exact (round (- x2 x1))))
                  (d (gcd up across))
                  (length (sqrt (+ (square (* up log-unit))
                                   (square (* across log-unit))))))
             (if (< length shortest-line)
                 (let ((x-inc (if (< across 0) -1.0 1.0))
                       (y-inc (if (< up 0) -1.0 1.0)))
                   (do ((x (x-actual x1) (+ x x-inc))
                        (y (y-actual y1) (+ y y-inc))
                        (i (inexact->exact (truncate length)) (- i 1)))
                       ((= i 0))
                     (emit "\\put(~a,~a){\\rule[-~apt]{~apt}{~apt}}"
                           x y
                           (/ unit-length 2.0)
                           unit-length
                           unit-length)))
                 (emit "\\put(~a,~a){\\line(~a,~a){~a}}"
                       (x-actual x1)
                       (y-actual y1)
                       (/ across d)
                       (/ up d)
                       (abs (* log-unit across)))))) ; LaTeX magic
          (else
           (error "Only vertical or horizontal lines: " (list x1 y1 x2 y2)))))

  (preamble)
  (for-each (lambda (c)
              (case (car c)
                ((line) (apply line (cdr c)))
                ((text-left) (apply text-extending-left (cdr c)))
                ((text-right) (apply text-extending-right (cdr c)))
                ((text-down) (apply text-extending-down (cdr c)))
                (else ???)))
            ((plot '*get)))
  (postamble))


; Generic latex table.
;
; out is an output-port.
;
; column-description is as for tabular, eg "|l|rr|rr|"
;
; header-description is a list of lists of fields.  Each sublist 
; describes one header row and constist of one field for each colum.
; Each field is either a string, or a list (string desc ...) where 
; desc can be a number (multicolumn field with, default 1) or 
; string (multicolumn format specifier, default is "c|").
;
; field-extractors is a list of functions, all are applied in 
; left-to-right order to the current row datum and the column number
; which is indexed from 0.  The extractor should return a datum, which 
; will be inserted verbatim into the output.  One space will be printed 
; on each side of the datum.
;
; Data is a list of row data, one row datum for each row to be printed.
;
; In order to support easy processing, some field extractors can be
; specified symbolically.  These assume that each row datum is a list
; and that the n'th extractor extracts and prints the value of the
; n'th datum in the list.
;  benchmark-name    formats a benchmark name
;  literal           just print it, unless #f
;  num-car           format car of datum as number, unless #f
;  num-cdr           format cdr of datum as number, unless #f
;  num-cadr          format cadr of datum as number, unless #f


(define (latex-table out
                     column-description
                     header-description
                     field-extractors
                     data)
  (define (header)
    (format out "\\begin{tabular}{~a}~%\\hline~%" column-description)
    (do ((i 0 (+ i 1))
         (h header-description (cdr h)))
        ((null? h))
      (do ((hf (car h) (cdr hf))
           (n 0 (+ n 1)))
          ((null? hf))
        (if (> n 0)
            (format out " & "))
        (let ((f (car hf)))
          (if (string? f)
              (format out "~a" f)
              (let loop ((x (cdr f)) (w 1) (spec "c|"))
                (cond ((null? x)
                       (format out "\\multicolumn{~a}{~a}{~a}"
                               w spec (car f)))
                      ((number? (car x))
                       (loop (cdr x) (car x) spec))
                      ((string? (car x))
                       (loop (cdr x) w (car x)))
                      (else ???))))))
      (format out " \\\\~%"))
    (format out "\\hline~%"))

  (define (format-conditional out format-spec d)
    (if d 
        (format out "~?" format-spec (list d))
        (format out "")))

  (define (format-symbolic f datum n format-spec)
    (cond ((eq? f 'benchmark-name)
           (format out "~?" format-spec 
                   (list (format-benchmark-name (list-ref datum n)))))
          ((eq? f 'num-car)
           (format-conditional out format-spec (car (list-ref datum n))))
          ((eq? f 'num-cdr)
           (format-conditional out format-spec (cdr (list-ref datum n))))
          ((eq? f 'num-cadr)
           (format-conditional out format-spec (cadr (list-ref datum n))))
          ((eq? f 'literal)
           (format-conditional out format-spec (list-ref datum n)))
          ((procedure? f)
           (format out "~?" format-spec (list (f datum n))))
          (else ???)))

  (define (body datum i)
    (for-each 
     (lambda (f n)
       (if (> n 0)
           (format out " & "))
       (cond ((procedure? f)
              (format out "~a" (f datum n)))
             ((symbol? f)
              (format-symbolic f datum n "~a"))
             ((pair? f)
              (format-symbolic (car f) datum n (cadr f)))
             (else ???)))
     field-extractors
     (iota (length field-extractors)))
    (format out " \\\\~%"))

  (define (footer)
    (format out "\\hline~%")
    (format out "\\end{tabular}~%"))

  (header)
  (let loop ((i 0) (data data))
    (if (not (null? data))
        (begin (body (car data) i)
               (loop (+ i 1) (cdr data)))))
  (footer))


; eof
