; Stop-and-copy results    -*- mode: scheme; mode: font-lock -*-
; 2000-09-11 / lth

(require 'control)

;----------------------------------------------------------------------

(define (canned-sc-tables-and-plots)
  (display "SC tables and plots")
  (newline)

  (load-stopcopy-results)

  ; These create HTML tables
  (create-stopcopy-summary-tables 'html)
  (merge-stopcopy-summary-html-tables)
  (create-stopcopy-gc-tables 'html)
  (merge-stopcopy-gc-html-tables)

  ; These create LaTeX source
  (create-stopcopy-summary-tables 'latex)
  (create-stopcopy-gc-tables 'latex)

  ; These create data files for GNUplot
  (create-stopcopy-gctime-vs-mark/cons-plot "plots/sc-summary-plots/mc-")
  (create-stopcopy-load-vs-mark/cons-plot "plots/sc-summary-plots/L-")

  'canned-sc-tables-and-plots)

;----------------------------------------------------------------------

; Crunch data (but do not compute any results -- include raw numbers only)
; and write crunched data to files.

(define (crunch-stopcopy-results)
  (for-each (lambda (fs)
              (let ((input (car fs))
                    (output (cadr fs)))
                (display "*** Processing ")
                (display input)
                (newline)
                (let ((x (crunch input)))
                  (call-with-output-file output
                    (lambda (out)
                      (write x out))))))
            '(("test/output/2000-04-06-sc.upper.1" "crunched/sc.upper.1")
              ("test/output/2000-04-06-sc.upper.5" "crunched/sc.upper.5")
              ("test/output/2000-04-28-sc.upper+lower.5" 
               "crunched/sc.upper+lower.5")
              ("test/output/2000-04-28-sc.upper+lower.1" 
               "crunched/sc.upper+lower.1")
              ("test/output/2000-04-07-sc.var.1" "crunched/sc.var.1")
              ("test/output/2000-04-07-sc.var.5" "crunched/sc.var.5"))))

; Data gathered August 2000 and later.

(define (crunch-stopcopy-extra-results)
  (for-each (lambda (fs)
              (let ((input (car fs))
                    (output (cadr fs)))
                (display "*** Processing ")
                (display input)
                (newline)
                (let ((x (crunch input)))
                  (call-with-output-file output
                    (lambda (out)
                      (write x out))))))
            '(; EARLEY
              ("test/output/2000-08-21-earley-sc.upper.5" 
               "crunched/sc-extra.upper.5")
              ("test/output/2000-08-21-earley-sc.upper+lower.5" 
               "crunched/sc-extra.upper+lower.5")
              
              ; PERM and GCOLD (note: GCOLD with broken RNG)

              ("test/output/2000-08-26-sc.upper.1"
               "crunched/sc-extra2.upper.1")
              ("test/output/2000-08-26-sc.upper+lower.1"
               "crunched/sc-extra2.upper+lower.1")
              ("test/output/2000-08-26-sc.var.1"
               "crunched/sc-extra2.var.1")

              ; GCOLD:10,1000 extra run because it timed out on the
              ; previous run...

              ("test/output/2000-09-04-sc.upper.1"
               "crunched/sc-extra3.upper.1")
              ("test/output/2000-09-04-sc.upper+lower.1"
               "crunched/sc-extra3.upper+lower.1")
              ("test/output/2000-09-04-sc.var.1"
               "crunched/sc-extra3.var.1")

              )))

;----------------------------------------------------------------------

; Load crunched data and compute results in preparation for generating
; reports and plots.

(define sc-results-loaded #f)

(define (load-stopcopy-results)

  (define (remove-load-2.0 data)
    (filter (lambda (x) 
              (not (= (cadddr (system.info x)) 2.0)))
            data))

  (if (not sc-results-loaded)
      (begin
        (set! sc-results-loaded #t)
        (for-each (lambda (x)
                    (let ((file (car x))
                          (variable (cadr x)))
                      (environment-set! (interaction-environment)
                                        variable
                                        (load-crunched-data file))))
                  '(("crunched/sc.upper.1" sc-upper-1)
                    ("crunched/sc.upper.5" sc-upper-5)
                    ("crunched/sc.upper+lower.1" sc-upper+lower-1)
                    ("crunched/sc.upper+lower.5" sc-upper+lower-5)
                    ("crunched/sc.var.1" sc-var-1)
                    ("crunched/sc.var.5" sc-var-5)))

        ; Remove 2.0 runs from UPPER results because they don't really make
        ; sense and only muddy the statistics overall.  Other collectors do
        ; not measure UPPER at 2.0.

        (set! sc-upper-1 (remove-load-2.0 sc-upper-1))
        (set! sc-upper-5 (remove-load-2.0 sc-upper-5))

        ; Add results of EARLEY

        (set! sc-upper-5 
              (append (remove-benchmark '(5 earley:12) sc-upper-5)
                      (load-crunched-data "crunched/sc-extra.upper.5")))
        (set! sc-upper+lower-5
              (append (remove-benchmark '(5 earley:12) sc-upper+lower-5)
                      (load-crunched-data "crunched/sc-extra.upper+lower.5")))

        ; Add results of PERM and GCOLD

        (set! sc-upper-1
              (append sc-upper-1 
                      (load-crunched-data "crunched/sc-extra2.upper.1")))
        (set! sc-upper+lower-1
              (append sc-upper+lower-1 
                      (load-crunched-data "crunched/sc-extra2.upper+lower.1")))
        (set! sc-var-1
              (append sc-var-1
                      (load-crunched-data "crunched/sc-extra2.var.1")))

        ; Add GCOLD:10,1000 runs

        (set! sc-upper-1
              (append (remove-benchmark '(1 gcold:25_0_10_1000_200) sc-upper-1)
                      (load-crunched-data "crunched/sc-extra3.upper.1")))
        (set! sc-upper+lower-1
              (append (remove-benchmark '(1 gcold:25_0_10_1000_200) sc-upper+lower-1)
                      (load-crunched-data "crunched/sc-extra3.upper+lower.1")))
        (set! sc-var-1
              (append (remove-benchmark '(1 gcold:25_0_10_1000_200) sc-var-1)
                      (load-crunched-data "crunched/sc-extra3.var.1")))
        #t)))

; ORDER MATTERS in the return from this procedure!

(define (sc-canonical-data)
  (load-stopcopy-results)
  (fluid-let ((no-warnings #t))
    `((,(basic-statistics sc-upper-1) 1)
      (,(basic-statistics sc-upper-5) 5)
      (,(basic-statistics sc-upper+lower-1) 1)
      (,(basic-statistics sc-upper+lower-5) 5)
      (,(basic-statistics sc-var-1) 1) 
      (,(basic-statistics sc-var-5) 5))))


(define (generic-stopcopy-summary-tables prefix variables stats1 stats5 type)
  (do ((variables variables (cdr variables)))
      ((null? variables))
    (let* ((var  (car variables))
           (s    (symbol->string var))
           (file (case type
                   ((latex) (string-append "tables/" prefix s ".tex"))
                   ((html)  (string-append "html-tables/" prefix s ".html"))
                   (else ???)))
           (n    (- (char->integer 
                     (string-ref s (- (string-length s) 1)))
                    (char->integer #\0)))
           (data (environment-get (interaction-environment) 
                                  var))
           (limit (not (string=? "var" 
                                 (let ((l (string-length s)))
                                   (substring s (- l 5) (- l 2)))))))
      (display var)
      (newline)
      (with-output-to-file file
        (lambda ()
          (cond ((= n 1)
                 (if limit
                     (stats1 data 'no-load/trigger)
                     (stats1 data 'no-heap-limit)))
                ((= n 5)
                 (if limit
                     (stats5 data 'no-load/trigger)
                     (stats5 data 'no-heap-limit)))
                (else ???)))))))

(define (create-stopcopy-summary-tables . rest)

  (define type (if (null? rest) 'latex (car rest)))

  (define (stats list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")         ms run-cpu-best)
             (("CPU GC")      ms run-cpu-gc-best)
             (("Num GCs")     n  run-garbage-collections-best)
             (("Marked")      mb ,(lambda (get present?)
                                    (stopcopy-mark-volume 
                                     get
                                     (benchmark-designator* get))))
             (("Heap" "used") mb bytes-max-heap)
             )
           'no-warnings 'break 9 type 'no-variance
           'no-fail
           'filter flatten-option-noise
           'sort sort-by-trigger
           'header-filter (lambda (bm collector heap aux)
                            (values (format-benchmark-name bm) "SC" heap 
                                    (if (null? aux) 
                                        #f
                                        (format #f "T=~a" (car aux)))))
           rest))

  (load-stopcopy-results)
  (generic-stopcopy-summary-tables
   ""
   '(sc-upper-1 sc-upper-5 
     sc-upper+lower-1 sc-upper+lower-5
     sc-var-1 sc-var-5)
   stats
   stats
   type))

(define (merge-stopcopy-summary-html-tables)
  (merge-summary-html-tables
   "html-tables/sc-general.html"
   "Stop-and-copy results | General"
    '("sc-upper+lower-1" "sc-upper+lower-5"
      "sc-upper-1" "sc-upper-5"
      "sc-var-1" "sc-var-5")))

(define (create-stopcopy-gc-tables . rest)

  (define type (if (null? rest) 'latex (car rest)))

  (define (stats1 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms cpu-best)
             (("CPU" "GC")        ms cpu-gc-best)
             (("CPU" "User")      ms cpu-user-best)
             (("CPU" "Sys.")      ms cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get present?)
                                        (- (get 'cpu-best) 
                                           (get 'cpu-gc-best))))
             (("Copy" "gc")       mb bytes-copied-best)
             (("Move" "los")      mb bytes-moved-best))
           'no-warnings 'break 8 type 'no-variance 'no-load/trigger
           rest))

  (define (stats5 list-of-systems . rest)
    (apply standard-report list-of-systems
           `((("CPU")             ms run-cpu-best)
             (("CPU" "GC")        ms run-cpu-gc-best)
             (("CPU" "User")      ms run-cpu-user-best)
             (("CPU" "Sys.")      ms run-cpu-system-best)
             (("CPU" "Mut.")      ms ,(lambda (get present?)
                                        (- (get 'run-cpu-best) 
                                           (get 'run-cpu-gc-best))))
             (("Copy" "gc")       mb run-bytes-copied-best)
             (("Move" "los")      mb run-bytes-moved-best))
           'no-warnings 'break 8 type 'no-variance 'no-load/trigger
           rest))

  (generic-stopcopy-summary-tables
   "gc-"
   '(sc-upper-1 sc-upper-5
     sc-upper+lower-1 sc-upper+lower-5
     sc-var-1 sc-var-5)
   stats1 
   stats5
   type))

(define (merge-stopcopy-gc-html-tables)
  (merge-summary-html-tables
   "html-tables/sc-gc.html"
   "Stop-and-copy results | GC"
    '("gc-sc-upper+lower-1" "gc-sc-upper+lower-5"
      "gc-sc-upper-1" "gc-sc-upper-5"
      "gc-sc-var-1" "gc-sc-var-5")))

(define (stopcopy-avg-mutator-time)

  (define results 
    (make-benchmark-associator (lambda () (list 0 0))))

  (define (accumulate data n)
    (if (and (all-iterations-succeeded? data)
             (load/trigger=? data 3.0))
        (let ((probe ((results 'lookup) (benchmark-designator data))))
          (set-car! (cdr probe)  (+ (car (cdr probe))
                                    (- (cdr (assq 'run-cpu-best data))
                                       (cdr (assq 'run-cpu-gc-best data)))))
          (set-car! (cddr probe) (+ (car (cddr probe)) 1)))))

  (define dec (decimals 2))

  (for-each-benchmark-result accumulate (sc-canonical-data))
  (map (lambda (x)
         (list (car x)
               (dec (/ (cadr x) (caddr x) 1000.))))
       ((results 'results))))

(define (stopcopy-avg-gc-time)

  (define results 
    (make-benchmark-associator (lambda () (list 0 0))))

  (define (accumulate data n)
    (if (and (all-iterations-succeeded? data)
             (load/trigger=? data 3.0))
        (let ((probe ((results 'lookup) (benchmark-designator data))))
          (set-car! (cdr probe)  (+ (car (cdr probe))
                                    (cdr (assq 'run-cpu-gc-best data))))
          (set-car! (cddr probe) (+ (car (cddr probe)) 1)))))

  (define dec (decimals 2))

  (for-each-benchmark-result accumulate (sc-canonical-data))
  (map (lambda (x)
         (list (car x)
               (dec (/ (cadr x) (caddr x) 1000.))))
       ((results 'results))))

(define (stopcopy-mark-volume get bm)   ; in bytes
  (if (= 1 (car bm))
      (+ (get 'bytes-copied-best)
         (get 'bytes-moved-best)
         (* (static-size bm) (get 'garbage-collections-best)))
      (+ (get 'run-bytes-copied-best)
         (get 'run-bytes-moved-best)
         (* (static-size bm) (get 'run-garbage-collections-best)))))

(define (stopcopy-gctime get bm)        ; in seconds
  (if (= 1 (car bm))
      (/ (get 'cpu-gc-best) 1000.0)
      (/ (get 'run-cpu-gc-best) 1000.0)))

(define (stopcopy-gctime-normalized get bm) ; milliseconds per megabyte
  (/ (* (stopcopy-gctime get bm) 1000.0)
     (/ (canonical-allocation bm) (* 1024 1024))))

(define (stopcopy-mark-cons-ratio get bm)
  (/ (stopcopy-mark-volume get bm) (canonical-allocation bm)))

(define (stopcopy-effective-load get bm) ; That is _load_, as in 1/L.
  (/ (profiled-peak-usage bm) (get 'bytes-max-heap)))

(define (stopcopy-effective-inverse-load get bm) ; That is L
  (/ (get 'bytes-max-heap) (profiled-peak-usage bm)))

(define (create-stopcopy-vs-mark/cons-report fn extract-numerator . rest)
  (let ((raw-results (if (null? rest)
                         (list sc-upper-1 sc-upper-5
                               sc-upper+lower-1 sc-upper+lower-5
                               sc-var-1 sc-var-5)
                         (car rest))))
    (create-vs-mark/cons-report 
     fn 
     (lambda (results)
       (extract-data results
                     extract-numerator
                     stopcopy-mark-cons-ratio))
     raw-results)))

(define (create-stopcopy-gctime-vs-mark/cons-report fn)
  (create-stopcopy-vs-mark/cons-report fn stopcopy-gctime))

(define (create-stopcopy-benchmark-2d-plot fn independent dependent . rest)
  (let ((raw-results (if (null? rest)
                         (list sc-upper-1 sc-upper-5
                               sc-upper+lower-1 sc-upper+lower-5
                               sc-var-1 sc-var-5)
                         (car rest))))
    (create-benchmark-2d-plot
     fn 
     (lambda (results)
       (extract-data results independent dependent))
     raw-results)))

(define (create-stopcopy-gctime-vs-mark/cons-plot prefix)
  (create-stopcopy-benchmark-2d-plot prefix
                                     stopcopy-gctime
                                     stopcopy-mark-cons-ratio))

(define (create-stopcopy-load-vs-mark/cons-plot prefix)
  (create-stopcopy-benchmark-2d-plot prefix
                                     stopcopy-effective-inverse-load
                                     stopcopy-mark-cons-ratio
                                     (list sc-upper-1 sc-upper-5
                                           sc-var-1 sc-var-5)))

(define (create-stopcopy-load-vs-gctime-plot prefix)
  (load-stopcopy-results)
  (create-stopcopy-benchmark-2d-plot prefix
                                     stopcopy-effective-inverse-load
                                     stopcopy-gctime-normalized
                                     (list sc-upper-1 sc-upper-5
                                           sc-var-1 sc-var-5)))


; Returns list of min, max, mean, median, std.dev.

(define (report-stopcopy-copy-speed)
  (let ((data (sort (mappend report-copy-speed (sc-canonical-data))
                    <)))
    `((min ,(car data))
      (max ,(car (last-pair data)))
      (mean ,(mean data))
      (median ,(median data))
      (std-dev ,(standard-deviation data)))))

(define (show-stopcopy-copy-speed)
  (histogram 
   70
   (accumulate =
     (sort (mappend report-copy-speed (sc-canonical-data)) <))))


; Cost of collection expressed in milliseconds per megabyte.

(define (sc-copy-cost)
  (load-stopcopy-results)
  (let ((r1 (basic-statistics sc-var-1))
        (r2 (basic-statistics sc-var-5)))
    (map (lambda (x)
           (cons (caar x) 
                 (exact->inexact (mean (map cadr x)))))
         (accumulate-unordered
          (lambda (a b)
            (equal? (car a) (car b)))
          (filter-map 
           (lambda (x)
             (if (all-iterations-succeeded? x)
                 (list (benchmark-designator x)
                       (/ (cdr (assq 'run-cpu-gc-best x))
                          (/ (+ (cdr (assq 'run-bytes-copied-best x))
                                (cdr (assq 'run-bytes-moved-best x)))
                             (* 1024 1024))))
                 #f))
           (append r1 r2))))))

; eof
