; Regression testing            -*- mode: scheme; mode: font-lock -*-
; 2000-06-29 / lth

; FIXME: this does not yet test most of stopcopy.sch, generational.sch,
; dof.sch, or rof.sch.  Many functions in those files are worth testing,
; but it's not always so easy since they produced LaTeX or other textual
; output in files.  A slight change to them--change call-with-output-file
; to call-with-output-medium--would allow me to at least obtain the textual
; output in a string in many cases.  This is done in rof.sch.

(define test-results '())
(define canonical-results '())
(define test-not-found 0)
(define stop-testing)

(define (test-the-system)
  (set! test-results '())
  (set! test-not-found 0)
  (set! canonical-results 
        (if (file-exists? "scheme/test-results.dat")
            (begin
              (display "Reading old results...")
              (newline)
              (call-with-input-file "scheme/test-results.dat" read))
            '()))
  (set! *bs-cache* '())
  (set! gen-results-loaded #f)
  (set! dof-results-loaded #f)
  (load-gen-results)
  (load-dof-results)
  (call-with-current-continuation
   (lambda (k)
     (set! stop-testing (lambda () (k #f)))
     (display "Testing...")
     (newline)
     (run-the-tests)
     (if (> test-not-found 0)           ; Save results if updated
         (begin
           (display "Writing new results...")
           (newline)
           (call-with-output-file "scheme/test-results.dat"
             (lambda (out)
               (write test-results out)
               (newline out))))))))

(define (test/check tag data)

  (define (equal? x y)
    (cond ((eqv? x y) #t)
          ((and (pair? x) (pair? y))
           (and (equal? (car x) (car y)) (equal? (cdr x) (cdr y))))
          ((and (string? x) (string? y))
           (or (string=? x y)
               (eqfail x y)))
          ((and (vector? x) (vector? y))
           (or (vector-equal? x y)
               (eqfail x y)))
          ((and (bytevector? x) (bytevector? y))
           (or (bytevector-equal? x y)
               (eqfail x y)))
          ((and (structure? x) (structure? y))
           (or ((structure-comparator) x y)
               (eqfail x y)))
          ((and (number? x) (number? y) (not (= x x)) (not (= y y)))
           #t) ; Both are NaN...
          (else 
           (eqfail x y))))

  (define (vector-equal? v1 v2)

    (define (v-equal-loop i)
      (cond ((< i 0) #t)
            ((equal? (vector-ref v1 i) (vector-ref v2 i))
             (v-equal-loop (- i 1)))
            (else #f)))

    (if (= (vector-length v1) (vector-length v2))
        (v-equal-loop (- (vector-length v1) 1))
        #f))

  (define (eqfail x y)
    (display "EQUAL failed for ") (newline)
    (display x) (newline)
    (display y) (newline)
    #f)

  (set! test-results (cons (cons tag data) test-results))
  (let ((probe (assq tag canonical-results)))
    (if probe
        (if (not (equal? (cdr probe) data))
            (test/error tag))
        (set! test-not-found (+ test-not-found 1))))
  (unspecified))

(define (test/error tag)
  (display "Test failed for tag: ")
  (display tag)
  (newline)
  (stop-testing))

(define (run-the-tests)
  ; crunch.sch
  (test/crunch)

  ; common.sch
  (test/common)

  ; common-reports.sch
  (test/common-reports)

  ; basic-statistics.sch
  (test/basic-statistics)

  ; process2.sch
  (test/process2/option-noise)
  (test/process2/select-option)
  (test/process2/select-desirable)

  ; comparative.sch
  (test/comparative)
  (test/comparative/extract-data)
  (test/comparative/compare-data)

  ; rof.sch
  (test/rof)

  (unspecified))

; The inputs are crunched data and therefore this serves as a test 
; on CRUNCH and on LOAD-CRUNCHED-DATA.

(define (test/crunch)
  (display "test/crunch") (newline)
  (test/check 'crunch/2gen 2gen-var-5)
  (test/check 'crunch/4dof 4dof-upper-1))

(define (test/common)
  (display "test/common") (newline)
  (test/check 'common/comparable-runs/1
              (map comparable-runs? 
                   (basic-statistics 2gen-var-5)
                   (basic-statistics 2gen-var-5)))
  (test/check 'common/comparable-runs/2
              (map comparable-runs? 
                   (basic-statistics 2gen-var-5)
                   (reverse (basic-statistics 2gen-var-5))))
  (test/check 'common/for-each-benchmark-result
              (let ((res '()))
                (for-each-benchmark-result
                 (lambda (datum n)
                   (set! res (cons (cons (car datum) n) res)))
                 (2gen-canonical-data))
                res))
  (test/check 'common/find-system
              (find-system (append (basic-statistics 4dof-var-1)
                                   (basic-statistics 3dof-var-1))
                           '((dof 4) variable 1.25)
                           'gcbench:5)))
  
(define (test/common-reports)
  (display "test/common-reports") (newline)
  (test/check 'common-reports/extract-data
              (extract-data (basic-statistics 2gen-var-5)
                            (lambda (get bm)
                              (list (get 'run-cpu-best)
                                    (get 'bytes-max-heap)))))
  (test/check 'common-reports/compute-and-accumulate-statistics
              (let ((raw-results (list 2gen-var-5))
                    (report (lambda (results)
                              (extract-data results
                                            (lambda (get bm)
                                              (get 'run-cpu-best))
                                            (lambda (get bm) 
                                              (get 'bytes-max-heap))))))
                (compute-and-accumulate-statistics 
                 report 
                 raw-results 
                 cons))))

(define (test/basic-statistics)
  (display "test/basic-statistics") (newline)
  (test/check 'basic-statistics/2gen (basic-statistics 2gen-var-5))
  (test/check 'basic-statistics/4dof (basic-statistics 4dof-upper-1)))

(define (test/process2/option-noise)
  (display "test/process2/option-noise") (newline)
  (test/check 'process2/option-noise/remove-2gen 
              (remove-option-noise (basic-statistics 2gen-var-5)))
  (test/check 'process2/option-noise/remove-4dof
              (remove-option-noise (basic-statistics 4dof-upper-1)))
  (test/check 'process2/option-noise/flatten-2gen 
              (flatten-option-noise (basic-statistics 2gen-var-5)))
  (test/check 'process2/option-noise/flatten-4dof 
              (flatten-option-noise (basic-statistics 4dof-upper-1))))

(define (test/process2/select-option)
  (display "test/process2/select-option") (newline)
  (test/check 'process2/select-option/no-full-gc
              (select-option '(-dof-fullgc-frequency 0) #t
                             (basic-statistics 4dof-upper-1)))
  (test/check 'process2/select-option/shadow-remset
              (select-option '-dof-no-shadow-remsets #f 
                             (basic-statistics 4dof-upper-1))))

(define (test/process2/select-desirable)
  (display "test/process2/select-desirable") (newline)
  (test/check 'process2/select-desirable
              (select-most-desirable-run-and-resort
               (flatten-option-noise (basic-statistics 4dof-upper-1))
               (lambda (a b)
                 (let ((x (assq 'run-cpu-gc-best a))
                       (y (assq 'run-cpu-gc-best b)))
                   (or (and x (not y))
                       (and x y (< (cdr x) (cdr y)))))))))

(define (test/comparative)
  (display "test/comparative") (newline)
  (test/check 'comparative/trend
              (comparative-trend 
               (list (basic-statistics 2gen-upper-1)
                     (basic-statistics 3gen-upper-1))
               (list (dof-compute-canonical-data 1dof-upper-1)
                     (dof-compute-canonical-data 2dof-upper-1)
                     (dof-compute-canonical-data 3dof-upper-1)
                     (dof-compute-canonical-data 4dof-upper-1))
               (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
               (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
               min))
  (test/check 'comparative/all
              (compare-across-all (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
                                  (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
                                  (all-2gen-canonical-data)
                                  (all-3dof-canonical-data)
                                  min))
  (test/check 'comparative/by-benchmark
              (compare-by-benchmark 
               (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
               (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
               (all-2gen-canonical-data)
               (all-3dof-canonical-data)
               min)))

(define (test/comparative/extract-data)
  (display "test/comparative/extract-data") (newline)
  (test/check 'comparative/extract-data/trend-data
              (extract-benchmark-trend-data
               (compare-by-benchmark 
                (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
                (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
                (all-2gen-canonical-data)
                (all-3dof-canonical-data)
                min)))
  (test/check 'comparative/extract-data/space-time-by-benchmark
              (extract-by-benchmark-space-time (2gen-canonical-data) 
                                               (4dof-canonical-data)))
  (test/check 'comparative/extract-data/by-benchmark
              (extract-by-benchmark (2gen-canonical-data) 
                                    (4dof-canonical-data)
                                    (lambda (get bm)
                                      (/ (get 'run-cpu-best) 1000.0))
                                    (lambda (get bm)
                                      (get 'bytes-max-heap)))))

(define (test/comparative/compare-data)
  (display "test/comparative/compare-data") (newline)
  (test/check 'comparative/compare-data/all-cpu
              (compare-all-by-cpu-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/all-cpu-gc
              (compare-all-by-cpu-gc-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/all-cpu-mutator
              (compare-all-by-cpu-mutator-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/all-space
              (compare-all-by-space 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/benchmarks-cpu
              (compare-benchmarks-by-cpu-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/benchmarks-cpu-gc
              (compare-benchmarks-by-cpu-gc-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/benchmarks-cpu-mutator
              (compare-benchmarks-by-cpu-mutator-time 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data)))
  (test/check 'comparative/compare-data/benchmarks-space
              (compare-benchmarks-by-space 
               (all-2gen-canonical-data)
               (all-4dof-canonical-data))))

(define (test/rof)
  (test/check 'rof/noncanonical
              (rof-noncanonical-results '-phase-detection '-np-remset-limit))
  (test/check 'rof/heap-space-counts
              (call-with-output-string gen-vs-rof-heap-space-counts))
  (test/check 'rof/mark-cons-analysis
              (call-with-output-string rof-mark/cons-analysis))
  (test/check 'rof/remset-analysis
              (call-with-output-string rof-remset-analysis))
  (test/check 'rof/remset-activity-analysis
              (call-with-output-string rof-remset-activity-analysis))
  (test/check 'rof/feature-evaluation
              (call-with-output-string rof-feature-evaluation-table)))

; eof
