; Author: Lars T Hansen
; Date  : 2000-08-18

; Nonportable: uses make-hashtable, hashtable-put!, hashtable-get.
; Easy enough to fix.  (also makes storage profiling harder.)

; Updates: Can add a new outer-query term
;
;   (store word <query>)
;
; which merges the records in the solution set into the set for
; the word.  Eg,
;
;   (store foo*bar (and foo bar))
;
; associates the result set for (and foo bar) with the new
; word foo*bar.  A word can be any symbol.  

(define (textsearch-benchmark-small)
  (textsearch "small"
              "sophocles.txt"
              '(gifts 
                (and gifts with)
                (and gifts (or with her))
                (or with her)
                (show death))))

(define (textsearch-benchmark-large)
  (textsearch "large"
              "kingjames.txt"
              '((or the (or a an))
                (or jesus (or (or death fish) (and the not)))
                (or (or potiphar king) (or lord solomon) (or david goliath)
                    (or job ruth))
                (and (or the (or a an))
                     (or jesus (or (or death fish) (and the not)))
                     (or (or potiphar king) (or lord solomon) 
                         (or david goliath) (or job ruth)))
                (or (or the the) (or the the) (or the the) (or the the) 
                    (or the the) (or the the) (or the the) 
                    (or the the) (or the the) (or the the) (or the the)
                    (or the the) (or the the))
                (or (or i he she it they them) (or us we our their mine your))
                )))

(define (textsearch tag input-file queries)

  (define db (make-db))
  
  (define (read-input)
    (call-with-input-file input-file
      (lambda (in)
        (do ((l (read-line in) (read-line in))
             (i 0 (+ i 1)))
            ((eof-object? l)
             (display i)
             (newline))
          (if (and (> i 0) (zero? (remainder i 1000)))
              (begin (display i) (display " ")))
          (db-insert-line! db (make-line l))))))

  ; outer-query ::= (and <query> ...) 
  ;               | (or <query> ...)
  ;               | (show <query>) 
  ;               | <symbol>
  
  (define (process-query query)
    (if (and (pair? query) (eq? (car query) 'show))
        (display-query query (run-query (cadr query)) #t)
        (display-query query (run-query query) #f)))
  
  (define (display-query query result show?)
    (display query)
    (display ":   ")
    (display (evector-length result))
    (newline)
    (if show?
        (do ((i 0 (+ i 1)))
            ((= i (evector-length result)))
          (display "  ")
          (display (line-text (evector-ref result i)))
          (newline))))
  
  (define (run-query query)
    (cond ((and (pair? query) (eq? (car query) 'and))
           (merge-conjunctive db (map run-query (cdr query))))
          ((and (pair? query) (eq? (car query) 'or))
           (merge-disjunctive db (map run-query (cdr query))))
          ((symbol? query)
           (or (db-lookup-word db (string-downcase (symbol->string query)))
               (make-evector)))
          (else
           (error "Illegal query " query))))

  (set! query process-query)
  (display "Reading ")
  (read-input)
  (display (db-serialno db))
  (display " lines read.")
  (newline)
  (run-benchmark (string-append "textsearch-" tag)
                 1
                 (lambda ()
                   (for-each process-query queries))
                 (lambda (result)
                   #t))
  #t)


(define (merge-disjunctive db evectors)

  (define (merge)
    (let* ((result   (make-evector))
           (n        (length evectors))
           (indices  (make-vector n 0))
           (lengths  (list->vector (map evector-length evectors)))
           (evectors (list->vector evectors))
           (large    (+ (db-serialno db) 1)))
      (let down ()
        (let ((smallest 
               (let across ((i 0) (smallest '()) (least-serialno large))
                 (cond ((= i n) 
                        smallest)
                       ((>= (vector-ref indices i) (vector-ref lengths i))
                        (across (+ i 1) smallest least-serialno))
                       (else
                        (let ((s (line-serialno 
                                  (evector-ref (vector-ref evectors i)
                                               (vector-ref indices i)))))
                          (cond ((= s least-serialno)
                                 (across (+ i 1)
                                         (cons i smallest)
                                         least-serialno))
                                ((< s least-serialno)
                                 (across (+ i 1) (list i) s))
                                (else
                                 (across (+ i 1) 
                                         smallest
                                         least-serialno)))))))))
          (if (null? smallest)
              result
              (let ((i (car smallest)))
                (evector-extend! result 
                                 (evector-ref (vector-ref evectors i)
                                              (vector-ref indices i)))
                (do ((s smallest (cdr s)))
                    ((null? s))
                  (vector-set! indices (car s) 
                                       (+ 1 (vector-ref indices (car s)))))
                (down)))))))
 
  (cond ((null? evectors)       (make-evector))
        ((null? (cdr evectors)) (car evectors))
        (else                   (merge))))

(define (merge-conjunctive db evectors)
  
  (define (merge)
    (let* ((result   (make-evector))
           (n        (length evectors))
           (indices  (make-vector n 0))
           (evectors (list->vector evectors))
           (first    (vector-ref evectors 0))
           (limit    (evector-length first)))
      (do ((i 0 (+ i 1)))
          ((= i limit) 
           result)
        (let* ((x      (evector-ref first i))
               (serial (line-serialno x)))
          (let across ((j 1))
            (cond ((= j n)
                   (evector-extend! result x))
                  (else
                   (let* ((m   (vector-ref indices j))
                          (v   (vector-ref evectors j))
                          (lim (evector-length v)))
                     (let down ((k m))
                       (if (< k lim)
                           (let ((lserial (line-serialno (evector-ref v k))))
                             (cond ((< lserial serial)
                                    (down (+ k 1)))
                                   ((= lserial serial)
                                    (vector-set! indices j k)
                                    (across (+ j 1)))
                                   (else
                                    (vector-set! indices j 
                                                 (max m (- k 1))))))))))))))))
  
  (cond ((null? evectors)       (make-evector))
        ((null? (cdr evectors)) (car evectors))
        (else                   (merge))))


; Database

;(define-record db    ; a database
;  (ht                ; a hash table mapping words to ordered evector of lines
;   serialno          ; a global serial number for lines in the database
;   ))

(define (make-db ht serialno) (vector '*db* ht serialno))
(define (db-ht db) (vector-ref db 1))
(define (db-serialno db) (vector-ref db 2))
(define (db-serialno-set! db sn) (vector-set! db 2 sn))

(define make-db
  (let ((make-db make-db))
    (lambda ()
      (make-db (make-hashtable string-hash assoc) 0))))

(define (db-insert-line! db line)
  (let ((ht (db-ht db))
        (n  (db-serialno db)))
    (db-serialno-set! db (+ n 1))
    (line-serialno-set! line n)
    (for-each (lambda (w)
                (let ((x (hashtable-get ht w)))
                  (if x
                      (evector-extend! x line)
                      (hashtable-put! ht w (evector line)))))
              (line-words line))))

; Return packed evector of lines containing word, in serial number order.

(define (db-lookup-word db word)
  (hashtable-get (db-ht db) word))

        
; Line

;(define-record line  ; a line of text
;  (words             ; a list of constituent words without repetitions
;   text              ; the entire line, as input
;   serialno          ; the serial number of the line in the database
;   ))

(define (make-line words text serialno) (vector '*line* words text serialno))
(define (line-words line) (vector-ref line 1))
(define (line-text line) (vector-ref line 2))
(define (line-serialno line) (vector-ref line 3))
(define (line-serialno-set! line sn) (vector-set! line 3 sn))

(define make-line 
  (let ((make-line make-line))
    (lambda (l)
      (make-line (string->words l) l 0))))

; Extensible-vector

;(define-record evector
;  (v                 ; storage
;   length            ; highest index in use + 1
;   init              ; object used for initializing new elements
;   ))

(define (make-evector v length init) (vector '*evector* v length init))
(define (evector-v e) (vector-ref e 1))
(define (evector-length e) (vector-ref e 2))
(define (evector-init e) (vector-ref e 3))
(define (evector-v-set! e x) (vector-set! e 1 x))
(define (evector-length-set! e x) (vector-set! e 2 x))

(define make-evector
  (let ((make-evector make-evector))
    (lambda rest
      (cond ((null? rest)
             (make-evector (make-vector 10 #f) 0 #f))
            ((null? (cdr rest))
             (make-evector (make-vector (car rest) #f) 0 #f))
            ((null? (cddr rest))
             (make-evector (make-vector (car rest) (cadr rest)) 0 (cadr rest)))
            (else
             (error "Too many arguments to make-evector."))))))

(define (evector . elts)
  (let ((ev (make-evector)))
    (for-each (lambda (x) (evector-extend! ev x)) elts)
    ev))

(define (evector-ref ev k)
  (let ((v (evector-v ev)))
    (if (< k (vector-length v))
        (vector-ref v k)
        (evector-init ev))))

(define (evector-set! ev k x)
  (let* ((v (evector-v ev))
         (limit (vector-length v)))
    (if (< k limit)
        (begin (vector-set! v k x)
               (evector-length-set! ev (max (evector-length ev) (+ k 1))))
        (let* ((new-length (max (+ k 10) (* limit 2)))
               (w (make-vector new-length (evector-init ev))))
          (do ((i 0 (+ i 1)))
              ((= i limit)
               (evector-v-set! ev w)
               (evector-set! ev k x))
            (vector-set! w i (vector-ref v i)))))))

(define (evector-extend! ev x)
  (evector-set! ev (evector-length ev) x))

; Utility functions

(define (string->words l) 
  (remove-adjacent-duplicates
   (sort (map string-downcase (string-split l char-alphabetic?))
         string<?)
   string=?))

(define (string-split s constituent?)
  (let ((limit (string-length s)))
    (let loop ((i 0) (words '()))
      (cond ((>= i limit) 
             (reverse words))
            ((constituent? (string-ref s i))
             (let ((start i))
               (let loop2 ((i (+ i 1)))
                 (if (and (< i limit) (constituent? (string-ref s i)))
                     (loop2 (+ i 1))
                     (loop (+ i 1) (cons (substring s start i) words))))))
            (else
             (loop (+ i 1) words))))))

(define (remove-adjacent-duplicates l same?)

  (define (loop x l)
    (cond ((null? l)
           (list x))
          ((same? x (car l))
           (loop x (cdr l)))
          (else
           (cons x (loop (car l) (cdr l))))))

  (cond ((null? l) l)
        ((null? (cdr l)) l)
        (else (loop (car l) (cdr l)))))

(define read-line 
  (let ((buffer (make-string 80)))
    (lambda (port)

      (define (loop p k limit)
        (let ((c (read-char p)))
          (cond ((eof-object? c)
                 (if (zero? k)
                     c
                     (substring buffer 0 k)))
                ((char=? c #\newline)
                 (substring buffer 0 k))
                (else
                 (string-set! buffer k c)
                 (let ((k (+ k 1)))
                   (if (= k limit)
                       (let ((new-buffer (make-string (* limit 2))))
                         (do ((i 0 (+ i 1)))
                             ((= i k))
                           (string-set! new-buffer i (string-ref buffer i)))
                         (set! buffer new-buffer)
                         (loop p k (* limit 2)))
                       (loop p k limit)))))))

      (loop port 0 (string-length buffer)))))

; eof
